
"""
    This module contains an upgraded version of the 
        primitive built-in dict from python.
    This module also contains a custom-built error 
        used only for the class Dict.
    
    Current state: in development.
"""

# python
from sys import getsizeof
# core package
from core.__collections.__dict.exceptions import *


class Dict(dict):
    def __init__(self, *arguments, **keywordArguments):
        self.dictionary = {}
        if len(arguments) == 1:
            if type(arguments[0]) != dict:
                raise TypeError
            self.dictionary.update(arguments[0])
            
        elif len(arguments) > 1:
            if not all(type(x) == dict for x in arguments):
                raise TypeError
                
            for argument in arguments:
                self.dictionary.update(argument)
                
        if keywordArguments != {}:
            temporary = {}
            for key, argument in keywordArguments.items():
                temporary.update({key: argument})
            self.dictionary.update(temporary)


    def update(self, *arguments, **keywordArguments):
        if not all(type(x) == dict for x in arguments):
            raise TypeError
        
        for argument in arguments:
            self.dictionary.update(argument)
 
        if keywordArguments != {}:
            temporary = {}
            for key, argument in keywordArguments.items():
                temporary.update({key: argument})
            self.dictionary.update(temporary)


    def is_empty(self):
        return self.dictionary == {}


    def copy(self):
        if self.isempty():
            raise EmptyDictError
        return Dict(self.dictionary.copy())


    def clear(self):
        self.dictionary.clear()



"""
    core/__numbers.py
    
    module designed especially for numbers
    but its not like numpy, numpy is god compared to this
    
    author: @alexzander
"""


def fixed_set_precision_str(real_number: float, precision: int):
    """ takes a @real_number and returns it with specified @precision
        
        return: str version of the number 
    """ 
    
    # validation
    if type(real_number) not in [float, int]:
        raise TypeError

    return "{:.{decimals}f}".format(real_number, decimals=precision)


def fixed_set_precision_float(real_number: float, precision: int):
    """ takes a @real_number and returns it with specified @precision
        
        return: float version of the number 
    """ 
    
    return float(fixed_set_precision_str(real_number, precision))




# python
from getpass import getuser
# core
from core.__path import *
# core/sounds
from core.sounds.paths import sounds_remote_path
from core.sounds.update import UpdateSounds


def create_remote_database(remote_path: str,
               original_remote_folder: str, 
               modified_remote_folder: str, 
               remote_sounds_json_path: str, 
               type: str):
    """ creation of sounds files and folder for the first time """
    # reseting everything
    if exists(remote_path):
        delete_folder(remote_path)
    # creating folders
    if not exists(original_remote_folder):
        os.makedirs(original_remote_folder)
    if not exists(modified_remote_folder):
        os.makedirs(modified_remote_folder)
    
    # copy the files from type/original to C:/Users/$username/sounds/$type/original
    sep = get_path_sep(__file__)
    original_folder = sep.join(__file__.split(sep)[:-1]) + f"/{type}/original"
    
    __sounds = os.listdir(original_folder)
    for s in __sounds:
        copy_file(original_folder + sep + s, original_remote_folder + sep + s, __print=True)
    
    # after creating the folders for the first time
    # also create the modified sounds
    # in order to be ready with everything
    UpdateSounds(
        original_remote_folder,
        modified_remote_folder,
        remote_sounds_json_path
    )


def check_remote_database(__type: str):
    sounds_folder = sounds_remote_path.format(username=getuser(), type=__type)

    # remote audio folders
    original_remote_folder = sounds_folder + "/original"
    modified_remote_folder = sounds_folder + "/modified"

    # remote database
    remote_sounds_json_path = sounds_folder + "/sounds.json"

    if  not exists(sounds_folder) or \
        is_folder_empty(sounds_folder) or \
        not exists(original_remote_folder) or \
        is_folder_empty(original_remote_folder) or \
        not exists(modified_remote_folder) or \
        is_folder_empty(modified_remote_folder) or \
        not exists(remote_sounds_json_path) or \
        is_file_empty(remote_sounds_json_path):
        
        # if one of the statements is true
        # then initialize everytime
        create_remote_database(
            sounds_folder,
            original_remote_folder, 
            modified_remote_folder,
            remote_sounds_json_path,
            __type
        )
    
    return remote_sounds_json_path



"""
    core/drive.py

    useful in development of programs that
    involve work with the disk drive or any
    disk drive

    author: @alexzander
"""


# python
import os
import string
import threading
import subprocess
import multiprocessing
from time import time, sleep
from concurrent.futures import ThreadPoolExecutor

# core package (pip install python-core)
from core.system import *
from core.aesthetics import *
from core.exceptions import *
from core.__winapi import *
from core.__numbers import *
from core.__path import *

# 3rd party
from win32 import win32api # pip install pywin32


def transfer__(src: str, dst: str):
    """
        first: content is validated
        second: reads from @src in binary and writes in binary to @dst
    """

    # validation
    if not exists(src):
        raise FileNotFoundError
    if not is_file(src):
        raise NotAFileError
    if not exists(dst):
        path = get_path_from_absolute(dst)
        # if the file @dst doesnt exist
        # the folders path is created
        # except the name of the file
        # that is created below in the
        # copy 'with' statement
        os.makedirs(path, exist_ok=True)
    # /validation

    with open(src, "rb") as src_bin:
        # if the @dst file doesnt exist
        # is created automatically
        with open(dst, "wb") as dst_bin:
            # deleted existed content
            dst_bin.truncate(0)
            # replaced with brand-new one
            dst_bin.write(src_bin.read())


def copy_file(src: str, dst: str, __print=False):
    """ copys from source to destination (available only for FILES)
        return True (means that code was executed successfully)

        prints output of code=0 to the screen if __print
    """

    # path validation
    if not is_file(src):
        raise NotAFileError
    if not is_valid_path(dst):
        raise InvalidPathError
    # path /validation

    sep_src = get_path_sep(src)
    sep_dst = get_path_sep(dst)

    # extensions validation
    filename_src = src.split(sep_src)[-1] # filename
    filename_dst = dst.split(sep_dst)[-1] # filename
    if filename_src != filename_dst:
        # if the same is diff, means
        # tha they have different extenions
        raise DifferentExtensionsError("{}\n{}".format(filename_src, filename_dst))
    # /extensions validation

    # @dst folder should exist in the first place
    # after that you can copy the file successfully
    try:
        transfer__(src, dst)
    except PermissionError as error:
        print(error.message)
        print(error)
        print(type(error))
        return False
    except Exception as error:
        print_red_bold("something was wrong")
        print_red_bold(error)
        print_red_bold(type(error))
        return False

    if __print:
        print()
        print("[" + "=" * 50 + "]")
        print("source: {}".format(blue(src)))
        print("[~]")
        print("dest: {}".format(cyan(dst)))
        print("[~]")
        print(green_bold("file copied successfully!"))
        print("[" + "=" * 50 + "]")
        print()

    # code=0
    return True


def copy_folder(src: str, dst: str, __print=False, __folderignore=[]):
    """ copys from @source_folder to @destination_folder (available only for FOLDERS)

        prints output of code=0 to the screen if __print
    """

    # validation
    if not is_folder(src):
        raise NotAFolderError
    if not is_valid_path(dst):
        raise InvalidPathError
    # /validation

    sep_src = get_path_sep(src)
    sep_dst = get_path_sep(dst)

    # iterating through __items in that folder
    for __item in os.listdir(src):
        src_path = src + sep_src + __item
        dst_path = dst + sep_dst + __item

        if is_file(src_path):
            copy_file(src_path, dst_path, __print)
        elif is_folder(src_path):
            if __folderignore:
                if "" in __folderignore:
                    __folderignore = list(filter(lambda e: e != "", __folderignore))
                if __item in __folderignore:
                    continue

            copy_folder(src_path, dst_path, __print, __folderignore)

            if __print:
                print()
                print("[" + "=" * 50 + "]")
                print("source: {}".format(blue(src_path)))
                print("[~]")
                print("dest: {}".format(cyan(dst_path)))
                print("[~]")
                print(green_bold("folder copied successfully!"))
                print("[" + "=" * 50 + "]")
                print()

    # code=0
    return True


def create_shortcut(src: str, dst: str, __print=False):
    """
        creates shortcut of @src and puts it into
        @dst with name

        prints output of code=0 to the screen if __print
    """
    # params validation
    if not is_file(src):
        raise NotAFileError
    if not is_folder(dst):
        raise NotAFolderError
    # /params validation

    sep_src = get_path_sep(src)
    sep_dst = get_path_sep(dst)

    # current working folder
    cwd = sep_src.join(src.split(sep_src)[:-1])

    delete_last_slash(dst)

    # creation of shortcut
    shell = Dispatch("WScript.Shell")
    shortcut = shell.CreateShortcut(dst + sep_dst + get_file_name(src) + ".lnk")
    shortcut.TargetPath = src
    shortcut.WorkingDirectory = cwd
    shortcut.IconLocation = src
    shortcut.save()

    if __print:
        print()
        print("[" + "=" * 50 + "]")
        print("source: {}".format(blue(src)))
        print("[~]")
        print("destination: {}".format(cyan(dst)))
        print("[~]")
        print(green_bold("was made shortcut successfully!"))
        print("[" + "=" * 50 + "]")
        print()

    # code=0
    return True


def is_flash_drive(drive: str):
    """ checks if a drive is USB-drive or NOT """
    try:
        os.listdir(drive)
        return True
    except PermissionError:
        return False


def get_available_drives():
    """ gets all available drives """
    ops = get_os()
    if ops == "Windows":
        return win32api.GetLogicalDriveStrings().split("\000")[:-1]
    elif ops == "Linux":
        return os.listdir("/media")
    elif ops == "Darwin":
        return os.listdir("/Volumes")


def get_available_drives_non_usb():
    """ return all NON-USB drives that you can copy files to """
    return [dv for dv in get_available_drives() if is_flash_drive(dv)]


def open_folder(path: str):
    """ opens folder """

    # validation
    if not is_folder(path):
        raise NotAFolderError
    # /validation

    ops = get_os()
    if ops == "Windows":
        subprocess.Popen(["explorer.exe", path])
    elif ops == "Linux":
        os.system("xdg-open \"{}\"".format(path))
    elif ops == "Darwin":
        os.system("open \"{}\"".format(path))

def open_file(path: str):
    """ opens a file """

    # validation
    if not is_file(path):
        raise NotAFileError
    # /validation

    ops = get_os()
    if ops == "Windows":
        os.startfile(path)
    elif ops == "Linux":
        os.system("xdg-open \"{}\"".format(path))
    elif ops == "Darwin":
        os.system("open \"{}\"".format(path))

def delete_folder(path: str):
    """ deletes a folder """

    if not is_folder(path):
        raise NotADirectoryError

    try:
        ops = get_os()
        if ops == "Windows":
            os.system("rmdir /S /Q {}".format(path))
        elif ops == "Linux":
            os.system("rm -r {}".format(path))
        elif ops == "Darwin":
            os.system("rm -rf {}".format(path))
        return True
    except PermissionError as error:
        # system folder or another process is using it
        print(error.message)
        print(type(error))
        print(error)
        return False
    except Exception as error:
        print(error.message)
        print(type(error))
        print(error)
        return False


def extract_folders(file: str, path: str, curr_lvl=1, deep_lvl=2, __print=False):
    """
        function used to gather folders on different
        deepness level in order to maximize efficient in
        file searching on disk

        return => @extracted_folders, list of folder paths
            that will be helpful for creating a thread
            searching on that folder path for @file

        raise StopRecursive exception if file is found
    """

    # list with paths
    extracted_folders = []
    try:
        sep = get_path_sep(path)
        for item in os.listdir(path):
            full_path = path + sep + item

            if is_folder(full_path):
                if __print:
                    print_red(full_path)
                if curr_lvl < deep_lvl:
                    extracted_folders.extend(extract_folders(file, full_path, curr_lvl + 1, deep_lvl, __print))
                elif curr_lvl == deep_lvl:
                    extracted_folders.append(full_path)

            elif is_file(full_path):
                if file == item:
                    if __print:
                        print_green("\nfile found!\n")
                        print("located on: {}".format(blue_underlined(full_path)))
                    raise StopRecursive(full_path)
                else:
                    if __print:
                        print_red(full_path)

    except NotADirectoryError:
        if file == path.split(sep)[-1]:
            if __print:
                print_green_bold("\nfile found!\n")
                print("located on: {}".format(blue_underlined(path)))
            raise StopRecursive(path)

    except PermissionError:
        pass

    return extracted_folders


def find_file_on_drive(file: str, path: str, __threading=True, deep_lvl=1, __print=False):
    """
        if @__threading is True
            this function search the @file on many threads concurrently
            no. of threads == no. of folder in the @path folder

            it is recommended to have as many folders as possible in the @path folder
            otherwise if you have 3 folders means 3 threads and if you are searching in 1 TB of data
                its gonna take a while


        else
            searches recursively on a single thread (not recommended)

        if file was found
            return file_path
        else
            return False

    """
    if __threading:
        search_tasks = []
        try:
            extracted_folders = extract_folders(file, path, deep_lvl=deep_lvl, __print=__print)
        except StopRecursive as result:
            # we found the file before
            # creating the search threads
            return result.message

        max_workers = len(extracted_folders)
        for path in extracted_folders:
            search_tasks.append((__find_file_on_drive, [file, path, __print]))

        if __print:
            print()
            for task in search_tasks:
                print(task)
            print("\nThere are [ {} ] threads prepared for execution.\n".format(red_bold(len(search_tasks))))

            choice = input("proceed? [y/n]:\n")
            if choice != "y":
                return red_bold("finding-process-canceled")

            print_yellow_bold("\nsearching...\n")

            before = None

        try:
            workers_results = []
            with ThreadPoolExecutor(max_workers=max_workers) as executor:
                if __print:
                    before = time()

                for task in search_tasks:
                    workers_results.append(executor.submit(task[0], *task[1]))

                for wr in workers_results:
                    wr = wr.result()
                    if wr != None and is_abs(wr):
                        if __print:
                            after = time() - before
                            after = fixed_set_precision_str(after, 2)
                            print("from thread-pool")
                            print("execution time: [ {} seconds(s) ]".format(yellow_bold(after)))
                        return wr

        except StopRecursive as result:
            if __print:
                after = time() - before
                after = fixed_set_precision_str(after, 2)
                print("from exception")
                print("execution time: [ {} seconds(s) ]".format(yellow_bold(after)))
            return result.message

    else:
        # non-threaded version (ignore)
        try:
            __find_file_on_drive(file, path, __print)
        except StopRecursive as result:
            return result.message

    # file was not found on the entire disk
    # case 1: it doesnt exist
    # case 2: file name is incorrectly provided
    return False


def __find_file_on_drive(file: str, path: str, __print=False):
    try:
        sep = get_path_sep(path)
        for item in os.listdir(path):
            full_path = path + sep + item

            if is_folder(full_path):
                if __print:
                    print_red(full_path)
                __find_file_on_drive(file, full_path, __print)

            elif is_file(full_path):
                if file == item:
                    if __print:
                        print_green(full_path)
                        print_green_bold("\n\nfile found!")
                        print("located on: {}\n\n".format(blue_underlined(full_path)))
                    raise StopRecursive(full_path)
                else:
                    if __print:
                        print_red(full_path)

    except NotADirectoryError:
        if file == path.split(sep)[-1]:
            print_green_bold("\nfile found!\n")
            print("located on: {}".format(blue_underlined(path)))

    except PermissionError:
        # this is a system folder
        # or another process is using this
        # file or folder
        pass

def get_size_in_bytes(path: str):
    size = 0
    if is_file(path):
        size += os.path.getsize(path)

    elif is_folder(path):
        sep = get_path_sep(path)
        for __item in os.listdir(path):
            size += get_size_in_bytes(path + sep + __item)

    return size


chunk_size = 1000
digital_units = ["KB", "MB", "GB", "TB", "PB", "EB", "ZB", "YB"]

def convert_size_in_bytes(size):
    """
        return tuple (
            size,
            digital measurement unit for size
        )
    """
    if size < chunk_size:
        return size, "bytes"

    index = -1
    while (size % chunk_size) != size:
        size //= chunk_size
        index += 1

    return size, digital_units[index]


def get_size_on_disk(path: str):
    """
        return tuple (
            size,
            digital measurement unit for size
        )
    """
    return convert_size_in_bytes(get_size_in_bytes(path))





# TESTING
if __name__ == '__main__':
    pass



"""
    core/download.py

    useful for downloading
    stuff from internet

    author: @alexzander
"""


# python
import json
import requests

# 3rd party
from bs4 import BeautifulSoup # pip install bs4
from requests_html import HTMLSession # pip install requests_html

# core package ( pip install python-core )
from core.__json import *


def download_image_from_url(url, destination):
    """ returns the absolute path of the destination """
    if type(url) != str or type(url) != str:
        raise TypeError("incorrect type.")
    if not url.startswith("http://") and not url.startswith("https://"):
        raise ValueError("not valid url.")
    if "/" not in destination and "\\" not in destination:
        raise ValueError

    response = requests.get(url, stream=True)
    response.raw.decode_content = True

    # try:
    #     with open(destination, "xb+") as binaryfile:
    #         binaryfile.write(response.raw.data)
    # except FileExistsError:
    with open(destination, "wb") as binaryfile:
        binaryfile.write(response.raw.data)
    return destination


def download_file_from_url(url, folder, headers={}, name="", extension=""):
    response = requests.get(url, headers=headers, stream=True)
    response.raw.decode_content = True

    if name == "" and extension == "":
        items = url.split("/")
        filename = items[len(items) - 1]
        fullpath = folder + "\\" + filename
        del items, filename
    else:
        if folder.endswith("\\"):
            fullpath = folder + name + "." + extension
        else:
            fullpath = folder + "\\" + name + "." + extension

    with open(fullpath, "wb") as binaryfile:
        binaryfile.truncate(0)
        binaryfile.write(response.raw.data)
    return fullpath


def get_yt_video_thumbnail_url(url):
    if type(url) != str:
        raise TypeError

    if not url.startswith("https://www.youtube.com/watch?v="):
        raise ValueError

    session = HTMLSession()
    response = session.get(url)
    response.html.render(sleep=1)

    soup = BeautifulSoup(response.html.html, "html.parser")
    webelement = soup.find("script", attrs={"class": "style-scope ytd-player-microformat-renderer"})

    data = json.loads(webelement.text)
    return data["thumbnailUrl"][0]




class Queue:
    def __init__(self, *objects):
        self.queue = []
        if len(objects) > 0:
            for obj in objects:
                self.queue.append(obj)


    def pop_front(self):
        self.queue = self.queue[1:]


    def append_back(self, *objects):
        for obj in objects:
            self.queue.append(obj)


    def front(self):
        if self.queue != []:
            return self.queue[0]

    def last(self):
        if self.queue != []:
            return self.queue[len(self.queue) - 1]


    def dimension(self):
        return len(self.queue)


    def clear(self):
        self.queue.clear()

    def isempty(self):
        return self.queue == []


    def __str__(self):
        return self.queue.__str__()


    def __repr__(self):
        return self.queue.__repr__()


    def __eq__(self, Queue_instance):
        return self.queue == Queue_instance.queue


# TESTING
if __name__ == '__main__':
    q = Queue(2, 3, 4, 5, 5, 6)
    print(q)
    
    q.pop_front()
    print(q)
    
    q.append_back()



"""
    core/gui/objects.py

    easy creation of tkinter object

    author: @alexzander
"""


# python
from tkinter import *


def NewFrame(root, position=TOP, expand=YES, fill=BOTH):
    frame = Frame(root)
    frame.pack(side=position, expand=expand, fill=fill)
    return frame


def NewButton(root, text, function, position=TOP, expand=YES, fill=BOTH):
    button = Button(root, text=text, command=function)
    button.pack(side=position, expand=expand, fill=fill)
    return button


def NewLabel(root, text, position=TOP, expand=YES, fill=BOTH):
    label = Label(root, text=text)
    label.pack(side=position, expand=expand, fill=fill)
    return label


def NewLabelWithBorder(root, text, position=TOP, borderwidth=2,
                       relief="solid", expand=YES, fill=BOTH):
    label = Label(root, text=text, borderwidth=borderwidth, relief=relief)
    label.pack(side=position, expand=expand, fill=fill)
    return label


def NewTextBox(root, text="", position=TOP, expand=YES, fill=BOTH, fore=None, back=None):
    textbox = Text(root, text=text)
    if fore:
        textbox.config(fore=fore)
    if back:
        textbox.config(back=back)
    textbox.pack(side=position, expand=expand, fill=fill)



"""
    core/algorithms.py

    powerful module for algo

    here are a lot of functions
    that you can use in your math programs

    author: @alexzander
"""


# python
import math


def greatest_common_divisor(x, y):
    while y:
        x, y = y, x % y
    return x


def absolute_value(real_number):
    if type(real_number) != float and type(real_number) != float:
        raise TypeError("param @real_number should be int or float.")

    return -real_number if real_number < 0 else real_number


def is_perfect_square(integer):
    if type(integer) != int:
        raise TypeError("param @integer should be int.")
    return math.sqrt(integer) == math.floor(math.sqrt(integer))


def is_prime(integer):
    if type(integer) != int:
        raise TypeError("param @integer should be int.")

    if integer < 2:
        return False
    elif integer > 2 and integer % 2 == 0:
        return False
    else:
        for divisor in range(3, int(math.sqrt(integer))):
            if integer % divisor == 0:
                return False
    return True


def factorial_recursive(integer):
    if type(integer) != int:
        raise TypeError("param @integer should be int.")

    if integer == 0:
        return 1
    return integer * factorial_recursive(integer - 1)


def factorial_iterative(integer):
    if type(integer) != int:
        raise TypeError("param @integer should be int.")

    computation = 1
    for value in range(2, integer):
        computation *= value

    return computation


def factorization(integer):
    """ returns a list with prime factors """
    if type(integer) != int:
        raise TypeError("param @integer should be int.")

    prime_factors = []
    while integer % 2 == 0:
        prime_factors.append(2)
        integer //= 2

    divisor = 3
    while divisor * divisor <= integer:
        while integer % divisor == 0:
            prime_factors.append(divisor)
            integer //= divisor

        divisor += 2
    if integer > 2:
        prime_factors.append(integer)

    return prime_factors


def reversed_number(integer):
    if type(integer) != int:
        raise TypeError("param @integer should be int.")

    reversed_number = 0
    while integer:
        reversed_number = reversed_number * 10 + integer % 10
        integer //= 10

    return reversed_number


def is_palindrome(integer):
    if type(integer) != int:
        raise TypeError("param @integer should be int.")

    return integer == reversed_number(integer)


def divisor_sum(integer):
    if type(integer) != int:
        raise TypeError("param @integer should be int.")

    import math

    computation = 0
    for divisor in range(1, int(math.sqrt(integer)) + 1):
        computation += divisor + integer // divisor

    if is_perfect_square(integer):
        computation -= math.sqrt(integer)

    return computation


def sieve_eratosthenes(dimension):
    if type(dimension) != int:
        raise TypeError("param @dimension should be int.")

    if dimension <= 1:
        raise ValueError("param @dimension cannot be <= 1")

    sieve = [1] * dimension
    sieve[0] = sieve[1] = 0

    import math
    for i in range(2, int(math.sqrt(dimension)) + 1):
        if sieve[i] == 1:
            for j in range(dimension + 1):
                expression = i ** 2 + j * i
                if expression < dimension:
                    sieve[expression] = 0
    primes = []
    for i in range(len(sieve)):
        if sieve[i]:
            primes.append(i)
    return primes


def power(base, exponent):
    if type(base) != int:
        raise TypeError("param @base should be int.")

    if type(exponent) != int:
        raise TypeError("param @exponent should be int.")

    if exponent == 0:
        return 1
    elif exponent < 0:
        return 1 / power(base, -exponent)

    calculation = 1
    for _ in range(exponent):
        calculation *= base
    return calculation


def compute_determinant(matrix):
    """ @matrix should be list of lists. """
    if type(matrix) != list:
        raise TypeError("param @matrix should be list of lists.")

    if len(matrix) == 1:
        return matrix[0][0]
    elif len(matrix) == 2:
        return matrix[0][0] * matrix[1][1] - matrix[0][1] * matrix[1][0]
    else:
        computation = 0
        for i in range(len(matrix)):
            minor_determinant = []
            for j in range(1, len(matrix)):
                minor_determinant.append([matrix[j][k] for k in range(len(matrix)) if k != i])

            computation += matrix[0][i] * (power(-1, i + 2)) * compute_determinant(minor_determinant)
        return computation


def anagram(str1ng, str2ng):
    if type(str1ng) != str:
        raise TypeError("param @w1 should be str.")

    if type(str2ng) != str:
        raise TypeError("param @w2 should be str.")

    frequency_str1ng = {}
    for character in str1ng:
        frequency_str1ng[character] = frequency_str1ng.get(character, 0) + 1

    frequency_str2ng = {}
    for character in str2ng:
        frequency_str2ng[character] = frequency_str2ng.get(character, 0) + 1

    return frequency_str1ng == frequency_str2ng


def is_corect_parantheses(pattern):
    """ "{[()[]()]}" returns True. """
    if type(pattern) != str:
        raise TypeError("param @pattern should be str.")

    stack = []
    mapped = {")":"(", "]":"[", "}":"{"}
    for character in pattern:
        if character in mapped:
            if stack:
                top_element = stack.pop()
            else:
                top_element = "#"

            if mapped[character] != top_element:
                return False
        else:
            stack.append(character)

    return True if stack else False


def quicksort(array, left, right, reverse=False):
    """ sorts the array in-place. """
    i = left
    j = right
    pivot = array[(left + right) >> 1]

    while i <= j:
        if reverse:
            while array[i] > pivot:
                i += 1
            while array[j] < pivot:
                j -= 1
        else:
            while array[i] < pivot:
                i += 1
            while array[j] > pivot:
                j -= 1
        if i <= j:
            array[i], array[j] = array[j], array[i]
            i += 1
            j -= 1

    if left < j:
        quicksort(array, left, j, reverse)
    if i < right:
        quicksort(array, i, right, reverse)


def quicksorted(array, reverse=False):
    """ returns a sorted copy of initial array. """
    if type(array) != list:
        raise TypeError("param @array should be list.")
    if type(reverse) != bool:
        raise TypeError("param @reverse should be bool.")

    new_array = array.copy()
    quicksort(array, 0, len(new_array) - 1, reverse)
    return new_array


def mergesort(array):
    if len(array) > 1:
        half = len(array) >> 1

        arr1 = array[:half]
        arr2 = array[half:]

        mergesort(arr1)
        mergesort(arr2)

        i = 0
        j = 0
        k = 0
        len_arr1 = len(arr1)
        len_arr2 = len(arr2)

        while i < len_arr1 and j < len_arr2:
            if arr1[i] < arr2[j]:
                array[k] = arr1[i]
                i += 1
            else:
                array[k] = arr2[j]
                j += 1
            k += 1

        for t in range(i, len_arr1):
            array[k] = arr1[t]
            k += 1

        for t in range(j, len_arr2):
            array[k] = arr2[t]
            k += 1


def mergesorted(array, reverse=False):
    new_array = array.copy()
    mergesort(new_array)
    if reverse:
        return list(reversed(new_array))
    return new_array


def binarysearch(array, element):
    """ returns index of found element """

    # validation
    if type(array) != list:
        raise TypeError("param @array should be list.")
    if type(element) != int:
        raise TypeError("param @element should be int.")

    left = 0
    right = len(array) - 1
    while left <= right:
        index = (left + right) >> 1
        if array[index] == element:
            return index
        elif array[index] > element:
            right = index - 1
        else:
            left = index + 1

    # no element found
    return -1


def binary_repr(self, integer):
    """ returns an aesthetic representaion of binary version of the transmited number. """

    # validation
    if type(integer) != int:
        raise TypeError("param @integer should be int.")

    binaryv = bin(integer)
    byte = [0 for _ in range(8)]
    index_byte = len(byte) - 1
    index_bin = len(binaryv) - 1

    for _ in range(len(binaryv)):
        if binaryv[index_bin] == "1":
            byte[index_byte] = 1
        elif binaryv[index_bin] == "b":
            break
        index_bin -= 1
        index_byte -= 1

    # print("Binary representation: {}(base 10) ----> {}(base 2)".format(number, byte))
    fill = ""
    result = ['' for _ in range(len(str(integer)))]
    for char in result:
        fill += char

    str_byte = "[ "
    for b in byte:
        str_byte += str(b) + " "
    str_byte += "]"

    indexes = [i for i in range(7, -1, -1)]
    str_indexes = "[ "
    for i in indexes:
        str_indexes += str(i) + " "
    str_indexes += "]"

    print("{:=^25}".format("BinaryRepr"))
    print("{} : {}".format(integer, str_byte))
    print("{} : {}".format(fill, str_indexes))
    print("=" * 25)


def subsets(instance):
    if type(instance) != set:
        raise TypeError("param @instance should be set.")

    import itertools
    result = []
    for counter in range(len(instance) + 1):
        combinations = itertools.combinations(instance, counter)
        for comb in combinations:
            result.append(set(comb))
    return result


def digit_to_word(digit):
    if type(digit) != int and type(digit) != str:
        raise TypeError("param @digit should be int or str.")

    if digit == "0" or digit == 0:
        return "zero"
    elif digit == "1" or digit == 1:
        return "one"
    elif digit == "2" or digit == 2:
        return "two"
    elif digit == "3" or digit == 3:
        return "three"
    elif digit == "4" or digit == 4:
        return "four"
    elif digit == "5" or digit == 5:
        return "five"
    elif digit == "6" or digit == 6:
        return "six"
    elif digit == "7" or digit == 7:
        return "seven"
    elif digit == "8" or digit == 8:
        return "eight"
    elif digit == "9" or digit == 9:
        return "nine"

    raise ValueError("a number with total digits bigger than 9 was given/")


def tenth_to_word(tenth):
    if type(tenth) != int and type(tenth) != str:
        raise TypeError("param @digit should be int or str.")
    if len(tenth) != 2:
        raise ValueError("param @tenth shuld have length of 2.")

    if tenth[0] == "1":
        if tenth[1] == "0":
            return "ten"
        elif tenth[1] == "1":
            return "eleven"
        elif tenth[1] == "2":
            return "twelve"
        elif tenth[1] == "3":
            return "thirteen"
        elif tenth[1] == "4":
            return "fourteen"
        elif tenth[1] == "5":
            return "fithteen"
        elif tenth[1] == "6":
            return "sixteen"
        elif tenth[1] == "7":
            return "seventeen"
        elif tenth[1] == "8":
            return "eighteen"
        elif tenth[1] == "9":
            return "nineteen"

    elif tenth[0] == "2":
        if tenth[1] == "0":
            return "twenty"
        else:
            return "twenty-" + digit_to_word(tenth[1])
    elif tenth[0] == "3":
        if tenth[1] == "0":
            return "thirty"
        else:
            return "thirty-" + digit_to_word(tenth[1])
    elif tenth[0] == "4":
        if tenth[1] == "0":
            return "forty"
        else:
            return "forty-" + digit_to_word(tenth[1])
    elif tenth[0] == "5":
        if tenth[1] == "0":
            return "fifty"
        else:
            return "fifty-" + digit_to_word(tenth[1])
    elif tenth[0] == "6":
        if tenth[1] == "0":
            return "sixty"
        else:
            return "sixty-" + digit_to_word(tenth[1])
    elif tenth[0] == "7":
        if tenth[1] == "0":
            return "seventy"
        else:
            return "seventy-" + digit_to_word(tenth[1])
    elif tenth[0] == "8":
        if tenth[1] == "0":
            return "eighty"
        else:
            return "eighty-" + digit_to_word(tenth[1])
    elif tenth[0] == "9":
        if tenth[1] == "0":
            return "ninety"
        else:
            return "ninety-" + digit_to_word(tenth[1])


def number_to_words(integer):
    if type(integer) != int:
        raise TypeError("param @interger should be int.")

    amplifiers = ["quadrillion", "trillion", "billion", "million", "thousand", ""]

    digits_rev = list(str(integer))[::-1]
    index = 3
    while index <= len(digits_rev):
        digits_rev.insert(index, "/")
        index += 4

    digits_rev.reverse()
    digits_rev = "".join(digits_rev).split("/")

    if digits_rev[0] == "":
        del digits_rev[0]

    print(digits_rev)
    number_name = ""
    for amp, group in zip(amplifiers[::-1], digits_rev[::-1]):
        if len(group) == 3:
            first = group[0]
            second = group[1]
            third = group[2]

            if first == "0" and second == "0" and third == "0":
                group_name = ""
            else:
                if first == "0" and second != "0" and third != "0":
                    group_name = tenth_to_word(second + third) + " " + amp
                elif first == "0" and second == "1" and third == "0":
                    group_name = tenth_to_word(second + third) + " " + amp
                elif first == "0" and second == "0" and third != "0":
                    group_name = digit_to_word(third) + " " + amp
                else:
                    group_name = digit_to_word(first) + " hundred"
                    if second == "0" and third != "0":
                        group_name += " " + digit_to_word(third)
                    else:
                        group_name += " " + tenth_to_word(second + third)
                    group_name += " " + amp
        elif len(group) == 2:
            group_name = tenth_to_word(group) + " " + amp
        elif len(group) == 1:
            group_name = digit_to_word(group) + " " + amp
        number_name = group_name + " " + number_name

    return number_name



"""
    core/statistics.py

    its obvious what it is for

    author: @alexzander
"""


def average(*arguments):
    if len(arguments) == 0:
        raise ValueError
    if len(arguments) == 1:
        if type(arguments[0]) != list:
            raise TypeError

        total_sum = 0
        for item in arguments[0]:
            total_sum += item
        return total_sum / len(arguments[0])

    total_sum = 0
    occurrences = 0
    for arg in arguments:
        if type(arg) == int:
            total_sum += arg
            occurrences += 1
        elif type(arg) == list:
            total_sum += sum(arg)
            occurrences += len(arg)
        else:
            raise TypeError
    return total_sum / occurrences


def median(*arguments):
    if len(arguments) == 0:
        raise ValueError
    if len(arguments) == 1:
        if type(arguments) != list:
            raise TypeError

        if len(arguments[0]) % 2 == 0:
            # midlle index
            i = len(arguments[0]) // 2
            return (arguments[0][i] + arguments[0][i - 1]) / 2
        else:
            # middle element
            return arguments[0][len(arguments[0]) // 2]

    for arg in arguments:
        if type(arg) != int:
            raise TypeError

    if len(arguments) % 2 == 0:
        # middle index
        i = len(arguments) // 2
        return (arguments[i] + arguments[i - 1]) / 2
    else:
        # middle element
        return arguments[len(arguments) // 2]


def mode(*arguments):
    if len(arguments) == 0:
        raise ValueError
    if len(arguments) == 1:
        if type(arguments[0]) != list:
            raise TypeError
        frequencies = {arg: arguments[0].count(arg) for arg in arguments[0]}
        most_frequent = max(frequencies.values())

        if most_frequent == 1:
            return None

        chosenones = [k for k, v in frequencies.items() if v == most_frequent]
        return tuple(chosenones)

    for arg in arguments:
        if type(arg) != int:
            raise TypeError

    frequencies = {arg: arguments.count(arg) for arg in arguments}
    most_frequent = max(frequencies.values())

    if most_frequent == 1:
        return None

    chosenones = [k for k, v in frequencies.items() if v == most_frequent]
    return tuple(chosenones)


def variance(*arguments):
    if len(arguments) == 0:
        raise ValueError
    if len(arguments) == 1:
        if type(arguments[0]) != list:
            raise TypeError

        _average = average(arguments[0])
        sumation = 0
        for value in arguments[0]:
            sumation += (value - _average) ** 2
        return sumation / (len(arguments[0]) - 1)

    for arg in arguments:
        if type(arg) != int:
            raise TypeError

    _average = average(arguments)
    sumation = 0
    for value in arguments:
        sumation += (value - _average) ** 2
    return sumation / (len(arguments) - 1)


def standard_deviation(*arguments):
    if len(arguments) == 0:
        raise ValueError
    if len(arguments) == 1:
        if type(arguments[0]) != list:
            raise TypeError
        import math
        return math.sqrt(variance(arguments[0]))

    for arg in arguments:
        if type(arg) != int:
            raise TypeError
    return math.sqrt(variance(*arguments))


def coeff_of_variation(*arguments):
    return standard_deviation(*arguments) + average(*arguments)


def covariance(leftset, rightset):
    varleft = variance(leftset)
    varright = 0
    meanright = average(rightset)
    for item in rightset:
        varright += (item - meanright) ** 2
    return (varleft * varright) / (len(leftset) - 1)


def correl_coeff(leftset, rightset):
    leftsd = standard_deviation(leftset)
    rightsd = standard_deviation(rightset)
    return covariance(leftset, rightset) / (leftsd * rightsd)



"""
    core/aesthetics.py

    useful and fancy design in terminal

    here you can
        - asciify text
        - color asciified text
        - color normal text

    author: @alexzander
"""


# python
import json
import requests

# core package ( pip install python-core )
from core.exceptions import *


impressive_fonts = [
    "bell",
    "big",
    "broadway",
    "bubble",
    "chunky",
    "contessa",
    "cursive",
    "cyberlarge",
    "cybermedun",
    "cybersmall",
    "digital",
    "doh",
    "doom",
    "double",
    "drpepper",
    "epic",
    "fender",
    "kban",
    "l4me",
    "larry3d",
    "ogre",
    "rectangles",
    "shadow",
    "slant",
    "small",
    "smkeyboard",
    "speed",
    "standard",
    "weird"
]


def asciify(text, font=None):
    """ puts 2 lines of spaces
         after the modified text. """
    if type(text) != str:
        text = str(text)
    if font not in impressive_fonts:
        raise NotFoundError

    text = "+".join(text.split())

    if font:
        url_parameters = f"/make?text={text}&font={font}"
    else:
        url_parameters = f"/make?text={text}"

    url = "http://artii.herokuapp.com"
    response = requests.get(url + url_parameters)
    if response.status_code != 200:
        raise ConnectionError("status code: {}".format(response.status_code))
    ascii_art = response.text
    items = ascii_art.split("\n")

    # deleting last 2 lines from the big text
    items = items[:len(items) - 2]
    # inserting "\n" because in original text it didnt exist
    items[len(items) - 1] += "\n"
    # reconstructing
    ascii_art = "\n".join(items)
    return ascii_art


def shift_left_ascii(ascii_text, size):
    """ left shift the big text with specified
        before:
        BIG_TEXT
        after:
        --------->(size) BIG_TEXT
    """
    if type(size) != int:
        raise TypeError

    if type(ascii_text) != str:
        raise TypeError

    if not "\n" in ascii_text:
        raise ValueError

    lines = ascii_text.split("\n")
    lines = [" " * size + line for line in lines]
    lines = "\n".join(lines)
    return lines


def shift_right_ascii(ascii_text, size):
    if type(size) != int:
        raise TypeError

    if type(ascii_text) != str:
        raise TypeError

    if not "\n" in ascii_text:
        raise ValueError

    lines = ascii_text.split("\n")
    lines = [line[size:] for line in lines]
    lines = "\n".join(lines)
    return lines


# text effects
endc_effect = "\033[0m"
bolded = "\033[1m"
underlined = "\033[4m"


ansi_codes = {
    'purple': '\033[95m',
    'blue': '\033[94m',
    'green': '\033[92m',
    'yellow': '\033[93m',
    'red': '\033[91m',
    'endc': '\033[0m',
    'bold': '\033[1m',
    'underlined': '\033[4m',
    'white': "\u001b[37;1m",
    "cyan": '\x1b[38;5;44m',
    "darkcyan": '\033[36m',
    "magenta": "\033[35m",
    "black": "\033[30m",
    "grey": "\x1b[38;5;246m",
    "orange": "\x1b[38;5;208m"
}


def ansi_colored(string, red, green, blue):
    if type(red) != int or \
       type(green) != int or \
       type(blue) != int or \
       type(string) != str:
        raise TypeError

    return "\x1b[{};{};{}m{}{}".format(red, green, blue, string, endc_effect)


def underlined(string):
    if type(string) != str:
        try:
            string = str(string)
        except:
            raise TypeError
    return ansi_codes["underlined"] + string + endc_effect


def bolded(string):
    if type(string) != str:
        try:
            string = str(string)
        except:
            raise TypeError
    return ansi_codes["bold"] + string + endc_effect


def ConsoleColored(string, color, bold=0, underlined=0):
    if type(string) != str:
        try:
            string = str(string)
        except Exception as error:
            print(error)
            message = red + ansi_codes["bold"] + 'type of param @string should be str.' + endc_effect
            raise TypeError(message)
            del message

    # incorrect color
    if color not in ansi_codes and color != 'random':
        message = red + ansi_codes["bold"] + 'this color "{}" is not in ANSICodesDICT.'.format(color) + endc_effect
        raise NotImplementedError(message)
        del message

    # bold == 0 and underlined == 0
    if not bold and not underlined:
        if color == "random":
            from random import choice
            return ansi_codes[choice(list(ansi_codes.keys()))] + string + endc_effect

        return ansi_codes[color] + string + endc_effect

    # bold == 0 and underlined == 1
    elif not bold and underlined:
        if color == "random":
            from random import choice
            return ansi_codes[choice(list(ansi_codes.keys()))] + \
                ansi_codes["underlined"] + string + endc_effect

        return ansi_codes[color] + ansi_codes["underlined"] + string + endc_effect

    # bold == 1 and underlined == 0
    elif bold and not underlined:
        if color == "random":
            from random import choice
            return ansi_codes[choice(list(ansi_codes.keys()))] + \
                ansi_codes["bold"] + string + endc_effect

        return ansi_codes[color] + ansi_codes["bold"] + string + endc_effect

    # bold == 1 and underlined == 1
    if color == "random":
        from random import choice
        return ansi_codes[choice(list(ansi_codes.keys()))] + \
            ansi_codes["bold"] + ansi_codes["underlined"] + string + endc_effect

    return ansi_codes[color] + ansi_codes["bold"] + ansi_codes["underlined"] + string + endc_effect


def print_ansi_table():
    import sys
    print("\n")
    for i in range(0, 16):
        for j in range(0, 16):
            code = str(i * 16 + j)
            sys.stdout.write(u"\u001b[38;5;" + code + "m " + code.ljust(4))
        print (u"\u001b[0m")
    print("\n")


import re

ansi_escape = re.compile(r"\x1B(?:[@-Z\\-_]|\[[0-?]*[ -/]*[@-~])")

def delete_ansi_escape_codes(string):
    if type(string) != str:
        try:
            string = str(string)
        except:
            raise TypeError

    return ansi_escape.sub("", string)


# ---------------------------------------------------------
# printing colored

def print_red(__string):
    print(ConsoleColored(__string, "red"))


def print_blue(__string):
    print(ConsoleColored(__string, "blue"))


def print_yellow(__string):
    print(ConsoleColored(__string, "yellow"))


def print_orange(__string):
    print(ConsoleColored(__string, "orange"))


def print_purple(__string):
    print(ConsoleColored(__string, "purple"))


def print_cyan(__string):
    print(ConsoleColored(__string, "cyan"))


def print_green(__string):
    print(ConsoleColored(__string, "green"))


# ---------------------------------------------------------
# printing colored UNDERLINED

def print_red_underlined(__string):
    print(ConsoleColored(__string, "red", underlined=1))


def print_blue_underlined(__string):
    print(ConsoleColored(__string, "blue", underlined=1))


def print_yellow_underlined(__string):
    print(ConsoleColored(__string, "yellow", underlined=1))


def print_orange_underlined(__string):
    print(ConsoleColored(__string, "orange", underlined=1))


def print_purple_underlined(__string):
    print(ConsoleColored(__string, "purple", underlined=1))


def print_cyan_underlined(__string):
    print(ConsoleColored(__string, "cyan", underlined=1))


def print_green_underlined(__string):
    print(ConsoleColored(__string, "green", underlined=1))


# ---------------------------------------------------------
# printing colored BOLD

def print_red_bold(__string):
    print(ConsoleColored(__string, "red", bold=1))


def print_blue_bold(__string):
    print(ConsoleColored(__string, "blue", bold=1))


def print_yellow_bold(__string):
    print(ConsoleColored(__string, "yellow", bold=1))


def print_orange_bold(__string):
    print(ConsoleColored(__string, "orange", bold=1))


def print_purple_bold(__string):
    print(ConsoleColored(__string, "purple", bold=1))


def print_cyan_bold(__string):
    print(ConsoleColored(__string, "cyan", bold=1))


def print_green_bold(__string):
    print(ConsoleColored(__string, "green", bold=1))


# ---------------------------------------------------------
# printing colored  BOLD + UNDERLINED

def print_red_bold_underlined(__string):
    print(ConsoleColored(__string, "red", bold=1, underlined=1))


def print_blue_bold_underlined(__string):
    print(ConsoleColored(__string, "blue", bold=1, underlined=1))


def print_yellow_bold_underlined(__string):
    print(ConsoleColored(__string, "yellow", bold=1, underlined=1))


def print_orange_bold_underlined(__string):
    print(ConsoleColored(__string, "orange", bold=1, underlined=1))


def print_purple_bold_underlined(__string):
    print(ConsoleColored(__string, "purple", bold=1, underlined=1))


def print_cyan_bold_underlined(__string):
    print(ConsoleColored(__string, "cyan", bold=1, underlined=1))


def print_green_bold_underlined(__string):
    print(ConsoleColored(__string, "green", bold=1, underlined=1))


# ---------------------------------------------------------
# returning colored

def yellow(__string):
    return ConsoleColored(__string, "yellow")


def orange(__string):
    return ConsoleColored(__string, "orange")


def purple(__string):
    return ConsoleColored(__string, "purple")


def cyan(__string):
    return ConsoleColored(__string, "cyan")


def green(__string):
    return ConsoleColored(__string, "green")


def red(__string):
    return ConsoleColored(__string, "red")


def blue(__string):
    return ConsoleColored(__string, "blue")


# ---------------------------------------------------------
# returning colored UNDERLINED

def yellow_underlined(__string):
    return ConsoleColored(__string, "yellow", underlined=1)


def orange_underlined(__string):
    return ConsoleColored(__string, "orange", underlined=1)


def purple_underlined(__string):
    return ConsoleColored(__string, "purple", underlined=1)


def cyan_underlined(__string):
    return ConsoleColored(__string, "cyan", underlined=1)


def green_underlined(__string):
    return ConsoleColored(__string, "green", underlined=1)


def red_underlined(__string):
    return ConsoleColored(__string, "red", underlined=1)


def blue_underlined(__string):
    return ConsoleColored(__string, "blue", underlined=1)


# ---------------------------------------------------------
# returning colored BOLD

def yellow_bold(__string):
    return ConsoleColored(__string, "yellow", bold=1)


def orange_bold(__string):
    return ConsoleColored(__string, "orange", bold=1)


def purple_bold(__string):
    return ConsoleColored(__string, "purple", bold=1)


def cyan_bold(__string):
    return ConsoleColored(__string, "cyan", bold=1)


def green_bold(__string):
    return ConsoleColored(__string, "green", bold=1)


def red_bold(__string):
    return ConsoleColored(__string, "red", bold=1)


def blue_bold(__string):
    return ConsoleColored(__string, "blue", bold=1)


# ---------------------------------------------------------
# returning colored BOLD + UNDERLINED

def yellow_bold_underlined(__string):
    return ConsoleColored(__string, "yellow", bold=1, underlined=1)


def orange_bold_underlined(__string):
    return ConsoleColored(__string, "orange", bold=1, underlined=1)


def purple_bold_underlined(__string):
    return ConsoleColored(__string, "purple", bold=1, underlined=1)


def cyan_bold_underlined(__string):
    return ConsoleColored(__string, "cyan", bold=1, underlined=1)


def green_bold_underlined(__string):
    return ConsoleColored(__string, "green", bold=1, underlined=1)


def red_bold_underlined(__string):
    return ConsoleColored(__string, "red", bold=1, underlined=1)


def blue_bold_underlined(__string):
    return ConsoleColored(__string, "blue", bold=1, underlined=1)


# TESTING
if __name__ == '__main__':
    print(asciify("hello"))



"""
    core/image.py

    this module is responsible for:
        - image to PNG
        - image to ICO
        - image to STRING
        (useful stuff in working with images)

    author: @alexzander
"""


# python
import binascii
import optparse
import pytesseract

# 3rd party
import numpy as np # pip install numpy
from PIL import Image # pip install Pillow

# core package (pip install python-core)
from core.__path import *
from core.system import *
from core.aesthetics import *


# you need to set up in your environment variables TESS = { path your tesseract.exe file }
if "TESS" not in os.environ.keys():
    raise NotFoundError
pytesseract.pytesseract.tesseract_cmd = os.environ["TESS"]


class image_Exception(Exception):
    def __init__(self, message=""):
        self.message = message


class NotAnImageError(image_Exception):
    """ file is not an image """
    pass


def rgb_2_hex(red: int, green: int, blue: int):
    try:
        red = int(red)
        green = int(green)
        blue = int(blue)
    except:
        raise ValueError("@red, @green or @blue cannot be converted to integer")
    return "#{:02x}{:02x}{:02x}".format(red, green, blue)


def decode_qr_bar_code(path):
    """ return Decoded class """

    # validation
    if type(path) != str:
        path = str(path)
    if not is_file(path):
        raise NotAFileError
    # /validation

    img = cv2.imread(path)
    decoded = pyzbar.decode(img)
    return decoded


def hex_2_rgb(hexadecimal: int):
    if type(hexadecimal) != str:
        raise TypeError

    # starting from 1 -> meaning that is excluding the '#'
    return map(ord, hexadecimal[1:].decode("hex"))


def image_to_str(path: str):
    """ return a string from image """
    return pytesseract.image_to_string(Image.open(path))


def save_pdf_from_image(path: str, dst_folder: str):
    """ saves pdf from image to dst specified by user """
    image_name = path.split("\\")[-1].split(".")[0]
    binary_pdf = pytesseract.image_to_pdf_or_hocr(Image.open(path))

    with open(dst_folder + "\\{}.pdf".format(image_name), "wb") as bin_file:
        bin_file.truncate(0)
        bin_file.write(binary_pdf)


def image_to_pdf(path: str):
    """ raw binary content of a pdf file from image file """
    return pytesseract.image_to_pdf_or_hocr(Image.open(path))


def image_to_png(path: str):
    path = get_path_from_absolute(path)
    filename = get_file_name(path)

    img = Image.open(path)
    img.save("{}/{}.png".format(path, filename))


def image_to_ico(path: str):
    img = Image.open(path)
    filename = get_file_name(path)
    path = get_path_from_absolute(path)
    img.save("{}/{}.ico".format(path, filename))

def save_gif(dst, images, fps=1):
    if type(images) != list:
        raise TypeError("param @images should be type list.")
    if type(dst) != str:
        raise TypeError("param @dst should be type str.")

    if not dst.endswith(".gif"):
        raise ValueError("param @destinaion should end with .gif extension.")

    import imageio
    images = [imageio.imread(image_abspath) for image_abspath in images]
    imageio.mimsave(dst, images, fps=fps)


def print_image(path):
    def get_ansi_color_code(r, g, b):
        if r == g and g == b:
            if r < 8:
                return 16
            if r > 248:
                return 231
            return round(((r - 8) / 247) * 24) + 232
        return 16 + (36 * round(r / 255 * 5)) + (6 * round(g / 255 * 5)) + round(b / 255 * 5)


    def get_color(r, g, b):
        return "\x1b[48;5;{}m \x1b[0m".format(int(get_ansi_color_code(r,g,b)))

    img = Image.open(path)
    print(img.size)

    height = 80
    width = int((img.width / img.height) * height)

    img = img.resize((width, height), Image.ANTIALIAS)
    img_array = np.asarray(img)
    print(img_array.shape)

    for h_index in range(height):
        for w_index in range(width):
            pix = img_array[h_index][w_index]
            print(get_color(pix[0], pix[1], pix[2]), sep='', end='')
        print()

# TODO
def hide_text_in_image(path, text):
    # hide text, trebuie sa vezi daca nu este encrypted deja
    pass

# TODO
def reveal_text_from_image(path):
    # reveal text, trebuie sa vezi daca exista ceva sa poti sa dai reveal
    pass


if __name__ == '__main__':
    pass




# python
import os

# core package ( pip install python-core )
from core.drive import *
from core.__json import *
from core.__path import *
from core.__audio import modify_volume


def UpdateSounds(original_remote_folder: str,
                 modified_remote_folder: str,
                 remote_sounds_json_path: str):
    """ 
        modifies every original sound with -15 dB and
            puts all the files into modified folder.
        
        if modified folder has items inside, will be
            deleted and replaced with brand new ones.
    """
    # verifying if is empty
    __items = os.listdir(modified_remote_folder)
    if __items != []:
        delete_folder(modified_remote_folder)
        # for file in __items:
        #     os.remove(modified_remote_folder + "/" + file)
    
    # saving all paths from original folder
    files = [
        original_remote_folder + "/" + file for file in os.listdir(original_remote_folder) if is_file((original_remote_folder + "/" + file))
    ]
    
    # -15 dB for every file in the original folder
    # -15 dB is the best volume you can get
    for f in files:
        modify_volume(f, modified_remote_folder, units=-15)
    
    # creating a json with paths for every modified sound
    sounds_modified_json = {}
    for file in os.listdir(modified_remote_folder):
        name = file.split("_")[0]
        sounds_modified_json[name] = modified_remote_folder + "/" + file
        
    # writing sounds json to disk in the remote location
    # creating sounds.json under:
    # C:/Users/$username/sounds/$type (windows)
    # or 
    # /home/$username/sounds/$type (linux)
    # or
    # /Users/$username/sounds/$type (mac)
    write_json_to_file(sounds_modified_json, remote_sounds_json_path)



"""
    core/__datetime.py

    wrapped around selenium

    author: @alexzader
"""


# 3rd party
from selenium import webdriver # pip install selenium

# core package ( pip install python-core )
from core.system import *


def init_chrome_driver(chrdv_path: str, user_agent: dict, icognito=True):
    options = webdriver.ChromeOptions()

    # only on windows
    __os = get_os
    if __os == "Windows":
        options.binary_location = "C:/Program Files (x86)/BraveSoftware/Brave-Browser/Application/brave.exe"
    elif __os == "Linux":
        options.binary_location = "/usr/bin/firefox/firefox-bin"
    elif __os == "Darwin":
        options.binary_location = "/Applications/Firefox.app/Contents/MacOS/firefox-bin"

    options.add_argument("user-agent={}".format(user_agent["User-Agent"]))
    if icognito:
        options.add_argument("--incognito")

    driver = webdriver.Chrome(executable_path=chrdv_path, options=options)
    return driver



"""
    graph.py
    
    useful graph collection
    
    author: @alexzander
"""


# python
import os

# core package
from core.system import *
from core.__json import *
from core.__collections.graph.exceptions import *
from core.__path import *
# from exceptions import * (same thing, it works)


class Graph:
    def __init__(self, edges_json: list):
        self.nodes = []
        self.graph_adj_dict = {}
        self.edges = list(map(tuple, edges_json))
        
        for edge in edges_json:
            if edge[0] in self.graph_adj_dict:
                self.graph_adj_dict[edge[0]].append(edge[1])
            else:
                self.graph_adj_dict[edge[0]] = [edge[1]]
                self.nodes.append(edge[0])


    def degree(self, node: int):
        for __node, adj_nodes in self.graph_adj_dict.items():
            if __node == node:
                return len(adj_nodes)
        raise VertexNotFoundError


    def get_adj_list_represenation(self):
        graph_representation = ""
        for index, vertex in enumerate(self.graph_adj_dict.keys()):
            if index == len(self.graph_adj_dict.keys()) - 1:
                graph_representation += "{} -> {}".format(vertex, self.graph_adj_dict[vertex])
            else:
                graph_representation += "{} -> {}\n".format(vertex, self.graph_adj_dict[vertex])
        return graph_representation


    def __str__(self):
        return self.get_adj_list_represenation()




class Node:
    def __init__(self, data):
        self.data = data
        self.right = None
        self.left = None


class BinaryTree:
    def __init__(self, root):
        self.root = Node(root)


    def preorder(self, start: Node, traversal=""):
        """ root => left => right"""
        if start:
            traversal += str(start.data) + "->"
            traversal = self.preorder(start.left, traversal)
            traversal = self.preorder(start.right, traversal)
        return traversal


    def inorder(self, start: Node, traversal=""):
        """ left => root => right"""
        if start:
            traversal = self.inorder(start.left, traversal)
            traversal += str(start.data) + "->"
            traversal = self.inorder(start.right, traversal)
        return traversal


    def postorder(self, start: Node, traversal=""):
        """ left => right => root"""
        if start:
            traversal = self.postorder(start.left, traversal)
            traversal = self.postorder(start.right, traversal)
            traversal += str(start.data) + "->"
        return traversal


# TESTING
if __name__ == '__main__':
    tree = BinaryTree(51)
    tree.root.left = Node(12)
    tree.root.right = Node(43)
    tree.root.left.left = Node(19)
    tree.root.left.right = Node(89)
    result = tree.preorder(tree.root)
    print(result)
    result = tree.inorder(tree.root)
    print(result)
    result = tree.postorder(tree.root)
    print(result)



"""
    core/__json.py

    upgraded version of json.py
    useful in work with jsons

    author: @alexzander
"""


# python
import os
import json

# core package ( pip install python-core )
from core.system import *
from core.__path import *


def json_dumps(__json: dict, __indent=4):
    return json.dumps(__json, indent=__indent)


def read_json_from_file(__path: str):
    """ reads .json from @path"""

    # validation
    if not is_file(__path):
        raise ValueError

    # read
    return json.loads(open(__path, "r+", encoding="utf-8").read())


def write_json_to_file(__json: dict, dst: str):
    """ writes dict or list from @dumped_json to @destination_path"""

    # validation
    # if not os.path.exists(destination_path):
    #     raise ValueError
    if type(__json) not in [dict, list]:
        raise TypeError

    # write
    with open(dst, "w+", encoding="utf-8") as _json:
        _json.truncate(0)
        _json.write(json.dumps(__json, indent=4))


def prettify(__collection, __indent=4):
    return json.dumps(__collection, indent=__indent)


def pretty_print(__collection, __indent=4):
    print(prettify(__collection, __indent=__indent))


# TESTING
if __name__ == '__main__':
    pass



"""
    core/pdf.py

    helpful with pdfs

    author: @alexzander
"""


# python
import os
from string import ascii_letters, digits, punctuation

# 3rd party
import pdfplumber # pip install pdfplumber
from PIL import Image # pip install Pillow
from PyPDF2.merger import PdfFileMerger # pip install pypdf2

# core package (pip install python-core)
from core.system import *
from core.aesthetics import ConsoleColored, underlined


def has_image_extension(img):
    ext = img.split(".")[1]
    image_extensions = ["jpg", "jpeg", "png", "JPG", "JPEG", "PNG"]
    if ext in image_extensions:
        return True
    return False


def get_text_from_pdf(__pdfpath: str):
    human_readable_chars = ascii_letters + digits + punctuation
    human_readable_chars = list(human_readable_chars) + ["\n", "\t", " "]

    __pdftext = ""
    with pdfplumber.open(__pdfpath) as pdf:
        for page in pdf.pages:
            modified_text = ""
            for char in page.extract_text():
                if char in human_readable_chars:
                    modified_text += char
            __pdftext += modified_text
    return __pdftext


def python_print_to_pdf(images_paths, PDFName):
    """
        alludes to: 'microsoft print to pdf' XDD
    """
    print("PythonPrintToPDF: processing...")

    # list with different paths with images
    if type(images_paths) == list:
        folder_path = images_paths[0].split("\\")
        folder_path = "\\".join(folder_path[:len(folder_path) - 1])
        ImagesArray = [Image.open(img).convert("RGB") for img in images_paths]

        print(ConsoleColored("merging photo(s) into pdf...", "yellow", bold=1))

        ImagesArray[0].save(folder_path + f"\\{PDFName}.pdf", save_all=True, append_images=ImagesArray[1:])
        location = folder_path

    # string parameter
    elif type(images_paths) == str:

        # folder with images
        if os.path.isdir(images_paths):
            pdf_path = images_paths + "\\{}.pdf".format(PDFName)

            if os.path.isfile(pdf_path):
                print(ConsoleColored("warning: pdf already existent.", "red", bold=1))
                os.remove(pdf_path)
                print(ConsoleColored("Existent pdf was deleted.", "red", bold=1))

            del pdf_path

            photos = [file for file in os.listdir(images_paths) if os.path.isfile(images_paths + "\\" + file) and has_image_extension(file)]

            if photos == []:
                raise ValueError(ConsoleColored("folder has no photos.", "red", bold=1))

            imgs = [images_paths + "\\" + p for p in photos]
            ImagesArray = [Image.open(img).convert("RGB") for img in imgs]

            print(ConsoleColored("merging photo(s) into pdf...", "yellow", bold=1))

            ImagesArray[0].save(images_paths + f"\\{PDFName}.pdf", save_all=True, append_images=ImagesArray[1:])
            location = images_paths

        # only one image
        elif os.path.isfile(images_paths):
            folder_path = images_paths.split("\\")
            folder_path = "\\".join(folder_path[:len(folder_path) - 1])
            image = Image.open(images_paths).convert("RGB")

            print(ConsoleColored("merging photo(s) into pdf...", "yellow", bold=1))

            image.save(folder_path + f"\\{PDFName}.pdf")
            location = folder_path

    PDFName += ".pdf"
    successfully = ConsoleColored("successfully", "green", bold=1)
    print("pdf with name {} created {}.".format(ConsoleColored(PDFName, "cyan", bold=1, underlined=1), successfully))
    location += "\\" + PDFName

    print("located on: {}".format(ConsoleColored(location, "yellow", bold=1, underlined=1)))
    os.system(location)


def PDFsMerger(pathx: str, pathy: str, dest_folder: str):
    merger = PdfFileMerger()
    merger.append(pathx)
    merger.append(pathy)

    name_pathx = pathx.split("\\")[-1].split(".")[0]
    name_pathy = pathy.split("\\")[-1].split(".")[0]

    merger.write("{}\\{}_{}_merged.pdf".format(dest_folder, name_pathx, name_pathy))
    merger.close()




import json
from hashlib import sha256
from collections import OrderedDict


class Block:
    def __init__(self, name, index, previous_hash=None):
        self.name = name
        self.index = index
        self.previous_hash = previous_hash
        self.current_hash = sha256(json.dumps(self.__dict__, indent=4).encode()).hexdigest()


    def __str__(self):
        return json.dumps({
            "name": self.name,
            "index": self.index,
            "prev_hash": self.previous_hash,
            "current_hash": self.current_hash
        }, indent=4)


if __name__ == '__main__':
    first_block = Block("first", 0)
    second_block = Block("second", 1, first_block.current_hash)
    third_block = Block("third", 2, second_block.current_hash)
    print(first_block, second_block, third_block, sep="\n")



"""
    core/__zip.py

    wrapped around zipfile.py

    author: @alexzander
"""


import os
import zipfile


def unzip(path, folder):
    if not os.path.isfile(path):
        raise ValueError
    if not os.path.isdir(folder):
        raise ValueError

    with zipfile.ZipFile(path, "r") as file:
        file.extractall(folder)


def zip_files(folder, zipname):
    if type(folder) == list:
        for file in folder:
            if not os.path.isfile(file):
                raise ValueError
    elif type(folder) == str:
        if not os.path.isfile(folder):
            raise ValueError
        folder = [folder]

    sep = "/" if "/" in folder[0] else "\\"
    working_directory = folder[0].split(sep)
    working_directory = sep.join(working_directory[:len(working_directory) - 1])


    wd_name = working_directory.split(sep)
    wd_name = wd_name[len(wd_name) - 1]
    for file in folder:
        items = file.split(sep)
        folder_name = items[len(items) - 2]
        if folder_name != wd_name:
            print(folder_name)
            print(working_directory)
            raise ValueError("files should be from the same directory.")

    originalWD = os.getcwd()
    os.chdir(working_directory)

    items = [file.split(sep) for file in folder]
    # just the filename
    folder = [item[len(item) - 1] for item in items]

    with zipfile.ZipFile("{}.zip".format(zipname), "w") as filesZIP:
        for file in folder:
            filesZIP.write(file)
    print("{}.zip created successfully.".format(zipname))
    return working_directory + "\\" + zipname + ".zip"


def zip_folder(folder, zipname):
    if not os.path.isdir(folder):
        raise ValueError

    originalWD = os.getcwd()
    os.chdir(folder)
    files = os.listdir(folder)

    with zipfile.ZipFile("{}.zip".format(zipname), "w") as file_zip:
        for file in files:
            file_zip.write(file)
    os.chdir(originalWD)

    print("{}.zip created successfully.".format(zipname))
    return folder + "\\" + zipname + ".zip"



"""
    system.py

    useful module in accesing and using system stuff

    author: @alexzander
"""


# python
import os
import sys
import platform
from time import sleep, time

# core package ( pip install python-core )
from core.aesthetics import *
from core.exceptions import *
from core.__numbers import *

# 3rd party
import cv2 # pip install opencv-python
import pyzbar # pip install pyzbar
import colored_traceback # pip install colored_traceback


def get_os():
    return platform.system()


def get_python_path():
    """ return folder path """
    if os.environ["PYTHONPATH"].endswith(";"):
        return os.environ["PYTHONPATH"][:-1]
    return os.environ["PYTHONPATH"]


def get_tesseract_path():
    """ return executable of tesseract engine (if installed) """
    if "TESS" not in os.environ.keys():
        raise NotFoundError("TESS is not environment variables")
    return os.environ["TESS"]


def colorize_error():
    colored_traceback.add_hook(always=True)


def clear_screen():
    """ clears the shell screen depending on the opearting system """
    o = get_os()
    if o == "Windows":
        os.system("cls")
    elif o == "Linux" or o == "Darwin":
        os.system("clear")


def pause_program(message="[ press {} to continue ... ]".format(green_bold("<enter>"))):
    input(message)


def get_device_name():
    return platform.node()


def get_arhitecture():
    return platform.architecture()[0]


def get_python_version():
    return platform.python_version()


def get_python_implementation():
    return platform.python_implementation()


def get_processor():
    return platform.processor()


def get_python_interpreter_path():
    return sys.executable


# TESTING
if __name__ == '__main__':
    o = get_os()
    print(o)



"""

"""

# python
import os
from datetime import datetime

# 3rd party
import pytz # pip install pytz

# core package ( pip install python-core )
from core.__json import *
from core.__path import *


class timezone_Exception(Exception):
    def __init__(self, message=""):
        self.message = message


class TimezoneNotFoundError(timezone_Exception):
    pass


def get_current_timezone(location="Europe/Bucharest", time_format="%H:%M:%S"):
    """ @location: Continent/City"""
    if location not in pytz.all_timezones:
        raise TimezoneNotFoundError
    return datetime.now(pytz.timezone(location)).strftime(time_format)



"""
    core/animations.py

    interesting module which includes
    colored animations

    included stuff:
        - colored progress bar
        - colored xmas tree

    author: @alexzander
"""


# python
from time import sleep
from random import choice

# core package ( pip install python-core )
from core.__numbers import *
from core.aesthetics import *
from core.system import clear_screen

# 3rd party
from colorama import Fore, Style


def progress_bar(iteration, length, decimals=2,
                 progressbar_length=40, fill_symbol="█",
                 color="yellow"):
    """
        [Progress]: |███████████████████████_________________| [58.00] [Complete]
    """
    if iteration > length:
        raise ValueError("iteration({}) value is greater than total_length({})".format(iteration, length))

    progress_percent = 100 * (iteration / float(length))
    progress_percent = fixed_set_precision_str(progress_percent, decimals)
    filled_length = int(progressbar_length * iteration // length)
    completed = fill_symbol * filled_length + "_" * (progressbar_length - filled_length)

    if iteration == length:
        progressbar = f"[Progress]: |{completed}| [{progress_percent}] [Completed]"
        progressbar = ConsoleColored(progressbar, "green")
    else:
        progressbar = f"[Progress]: |{completed}| [{progress_percent}] [Complete]"
        progressbar = ConsoleColored(progressbar, color)

    return progressbar


def load_progress_bar(length=100, sleep_duration=0.045, color="yellow"):
    for i in range(length + 1):
        p = progress_bar(i, length, color=color)
        print(p, end="\r")
        sleep(sleep_duration)
    print()


def xmas_tree(message, allocated_space=None, __iterations=True, pause=0.8):
    """
        example of animation wit @message="xmas_tree"
                       *
                       x
                      xma
                     xmas_
                    xmas_tr
                   xmas_tree
                     ||||
                     ||||
                 =============
                 =============
    """
    __len_message = len(message)
    if allocated_space == None:
        allocated_space = __len_message + 4
    else:
        if type(allocated_space) == int:
            if allocated_space <= __len_message:
                allocated_space = __len_message + 4
        else:
            raise TypeError

    tree_colors = [Fore.GREEN, Fore.BLUE, Fore.RED, Fore.YELLOW, Fore.WHITE, Fore.CYAN]

    if __iterations == True:
        while True:
            clear_screen()
            print(Style.BRIGHT + tree_colors[3] + "{:^{space}}".format("*", space=allocated_space))
            for i in range(1, __len_message + 1, 2):
                if i == __len_message:
                    print(Style.BRIGHT + choice(tree_colors) + "{:^{space}}".format(message, space=allocated_space))
                else:
                    print(Style.BRIGHT + choice(tree_colors) + "{:^{space}}".format(message[:i], space=allocated_space))

            print("{:^{space}}".format("||||", space=allocated_space))
            print("{:^{space}}".format("||||", space=allocated_space))
            print('=' * allocated_space)
            print('=' * allocated_space)
            sleep(pause)
        print(Style.RESET_ALL, end="")

    elif type(__iterations) == int and __iterations > 0:
        for _ in range(__iterations):
            clear_screen()
            print(Style.BRIGHT + tree_colors[3] + "{:^{space}}".format("*", space=allocated_space))
            for i in range(1, __len_message + 1, 2):
                if i == __len_message:
                    print(Style.BRIGHT + choice(tree_colors) + "{:^{space}}".format(message, space=allocated_space))
                else:
                    print(Style.BRIGHT + choice(tree_colors) + "{:^{space}}".format(message[:i], space=allocated_space))

            print("{:^{space}}".format("||||", space=allocated_space))
            print("{:^{space}}".format("||||", space=allocated_space))
            print('=' * allocated_space)
            print('=' * allocated_space)
            sleep(pause)
        print(Style.RESET_ALL, end="")





# TESTING
if __name__ == '__main__':
    xmas_tree("xmas_tree", __iterations=5)



"""
    core/exceptions.py

    useful custom exceptions for
    different cases

    author: @alexzander
"""


# core package ( pip install python-core )
from core.system import *
from core.aesthetics import *


class core_Exception(Exception):
    def __init__(self, message=""):
        self.message = message


class RuntimeError(core_Exception):
    """ error during execution """
    pass


class HTTP_RequestError(core_Exception):
    """ handles the http stuff"""
    pass


class NotFound_404_Error(HTTP_RequestError):
    """ 404 http error """
    pass


class Forbidden_403_Error(HTTP_RequestError):
    """ 403 http error """
    pass


class StupidCodeError(core_Exception):
    pass


class StopRecursive(core_Exception):
    pass


class NotFoundError(core_Exception):
    pass



"""

    core/__random.py

    useful random stuff to import in your scripts
    contents (random):
        lowchar
        upperchar
        digit
        number
        lowerstring
        upperstring
        string
        date str
        date struct

    author: @alexzander
"""



# python
import time
from datetime import datetime
from random import randint, choice, random


# unicode codes
# a-z 97-122
# A-Z 65-90
# 0-9 48 57
random_lowerchar = lambda : chr(randint(97, 122))
random_upperchar = lambda : chr(randint(65, 90))
random_digit = lambda : randint(0, 9)


def random_digits(dimension):
    if dimension is None:
        raise TypeError
    if dimension <= 0:
        raise ValueError
    return "".join([str(randint(0, 9)) for _ in range(dimension)])

def random_number(dimension):
    if dimension is None:
        raise TypeError
    if dimension <= 0:
        raise ValueError
    return int(str(randint(1, 9)) + "".join([str(randint(0, 9)) for _ in range(dimension - 1)]))

def random_lower_str(dimension):
    if dimension is None:
        raise TypeError
    return "".join([
        chr(randint(ord("a"), ord("z"))) for _ in range(dimension)
    ])

def random_upper_str(dimension):
    if dimension is None:
        raise TypeError
    return "".join([
        chr(randint(ord("A"), ord("Z"))) for _ in range(dimension)
    ])

def random_str(dimension):
    if dimension is None:
        raise TypeError
    return "".join([
        choice([
            chr(randint(ord("a"), ord("z"))),
            chr(randint(ord("A"), ord("Z")))
        ]) for _ in range(dimension)
    ])

def random_date_str(starting_date="01.01.1971", ending_date=datetime.now().strftime("%d.%m.%Y")):
    if starting_date is None or ending_date is None:
        raise TypeError

    datetime_format = "%d.%m.%Y"
    start_time = time.mktime(time.strptime(starting_date, datetime_format))
    stop_time = time.mktime(time.strptime(ending_date, datetime_format))
    random_time = start_time + random() * (stop_time - start_time)
    return time.strftime(datetime_format, time.localtime(random_time))

def random_date_struct(starting_date="01.01.1971", ending_date=datetime.now().strftime("%d.%m.%Y")):
    if starting_date is None or ending_date is None:
        raise TypeError

    datetime_format = "%d.%m.%Y"
    start_time = time.mktime(time.strptime(starting_date, datetime_format))
    stop_time = time.mktime(time.strptime(ending_date, datetime_format))
    random_time = start_time + random() * (stop_time - start_time)
    return time.strptime(time.strftime(datetime_format, time.localtime(random_time)), datetime_format)




class Stack:
    def __init__(self, *objects):
        self.stack = []
        if len(objects) > 0:
            for obj in objects:
                self.stack.append(obj)


    def push(self, *objects):
        for obj in objects:
            self.queue.append(obj)


    def pop(self, obj):
        if self.queue != []:
            self.stack.pop()


    def top_item(self):
        if self.queue != []:
            return self.stack[len(self.stack) - 1]


    def isempty(self):
        return self.stack == 0


    def dimension(self):
        return len(self.stack)


    def clear(self):
        self.stack.clear()


    def __str__(self):
        return self.stack.__str__()


    def __repr__(self):
        return self.stack.__repr__()


    def __eq__(self, Stack_instance):
        return self.stack == Stack_instance.stack


# usage
if __name__ == '__main__':
    q = Stack()
    q2 = Stack(1, 2, 3, 4, 5, 6, 7, 8)
    print(q)
    print(q2)

    q3 = Stack(1, 2, 3, 4, *[1, 2, 3, 4, 5], Stack(1, 2, 3, 4))
    print(q3)
    print(q == q3)



"""
    core/__audio.py

    helpful with development of programs
    that work with audio

    author: @alexzander
"""


# python
import os
import threading

# 3rd party
import pyttsx3 # pip install pyttsx3
import sounddevice # pip install sounddevice
from gtts import gTTS # pip install gtts
from pydub import AudioSegment # pip install pydub
from playsound import playsound # pip install playsound
import speech_recognition as sr # pip install SpeechRecognition
from scipy.io.wavfile import write # pip install scipy

# core package (pip install python-core)
from core.system import *
from core.__path import *
from core.aesthetics import *


def playaudio(path, __threaded=True):
    # validation
    if type(path) != str:
        raise TypeError("param @absolute_path should be str.")
    if not is_file(path):
        raise ValueError("param @absolute_path is not an absolute path.")
    if not path.endswith(".wav") and not path.endswith(".mp3"):
        raise ValueError("param @absolute_path should end with .wav or .mp3")
    # /validation

    if __threaded:
        threading.Thread(target=playsound, args=(path, )).start()
    else:
        playsound(path)


def save_text_to_speech(text: str, dst: str, lang="en", slowmotion=False, __print=False):
    # validation
    # @text
    if type(text) != str:
        text = str(text)
    if text == "":
        raise ValueError("param @text is void-string.")

    # @dst
    if type(dst) != str:
        raise TypeError("param @dst should be type str.")
    if not dst.endswith(".mp3") and not dst.endswith(".wav"):
        raise ValueError("param @dst should end with .mp3")
    if not is_abs(dst):
        raise ValueError

    # @lang
    if type(lang) != str:
        lang = str(lang)

    speech = gTTS(text, lang=lang, slow=slowmotion)
    speech.save(dst)

    if __print:
        message = "Text: '{}'\nwas saved to speech here: {}\n{}".format(yellow(text), cyan(dst), green_bold("successfully."))
        print(message)


def save_recorded_mic(dst, duration):
    # validation
    if type(dst) != str:
        raise TypeError("param @dst should be type str.")
    if type(duration) != int:
        raise TypeError("param @duration should be type int.")

    if not is_abs(dst):
        raise ValueError("param @dst should be absolute path.")

    if not dst.endswith(".wav"):
        raise ValueError("param @dst should end with .wav extension.")

    if duration <= 0:
        raise ValueError("param @duration should be bigger than 0.")
    # /validation

    try:
        sample_rate = 44100
        print("you have {} seconds to speak:".format(duration))
        print("recording...")

        myrecording = sounddevice.rec(int(duration * sample_rate), samplerate=sample_rate, channels=2)
        sounddevice.wait()

        write(dst, sample_rate, myrecording)

        print("recording saved to path: {}.".format(dst))

    except OSError as error:
        print(type(error))
        print(error)
        print("turn OFF your anti-virus application.")


def speech_to_text_from_mic():
    """ listens to your voice, if understands:
        return speech to text else return None.
    """
    recognizer = sr.Recognizer()
    microphone = sr.Microphone()

    print("listening...")
    with microphone as audio:
        recognizer.adjust_for_ambient_noise(audio)
        recorded_audio = recognizer.listen(audio)
    try:
        text = recognizer.recognize_google(recorded_audio)
    except sr.UnknownValueError:
        print("your speech was incomprehensible. None was returned.")
        return None
    return text


def speech_to_text_from_file(path):
    if not is_file(path):
        raise ValueError
    if not path.endswith(".mp3") and not path.endswith(".wav"):
        raise ValueError

    r = sr.Recognizer()
    with sr.AudioFile(path) as audio_file:
        data = r.record(audio_file)
        try:
            text = r.recognize_google(data)
            return text
        except sr.UnknownValueError as error:
            print(error)
            print(error.message)
            print(type(error))


def speak_engine(message, speech_speed=150):
    if type(message) != str:
        raise TypeError("param @message should be type str.")
    if message == "":
        raise ValueError("param @message shouldn't be void-string.")

    engine = pyttsx3.init("sapi5")
    engine.setProperty("rate", speech_speed)
    engine.setProperty("volume", 0.8)
    voices = engine.getProperty("voices")
    engine.setProperty("voice", voices[1].id)
    engine.say(message)
    engine.runAndWait()


def modify_volume(src, dst, units=15):
    """ applicable only to .wav and .mp3 files. """
    # validation
    if type(src) != str:
        message = ConsoleColored("param @absolute_path should be type str.", "red", bold=1)
        raise TypeError(message)
        del message
    if type(units) != int:
        message = ConsoleColored("param @quantity should be type int.", "red", bold=1)
        raise TypeError(message)
        del message

    if not os.path.isfile(src):
        message = ConsoleColored("param @absolute_path is not an absolut path.", "red", bold=1)
        raise ValueError(message)
        del message

    if not src.endswith(".wav") and not src.endswith("mp3"):
        message = ConsoleColored("param @absolute_path should end with .wav or .mp3.", "red", bold=1)
        raise ValueError(message)
        del message
    # /validation

    extension = get_file_extension(src)
    if extension == "wav":
        audio_file = AudioSegment.from_wav(src)
    elif extension == "mp3":
        audio_file = AudioSegment.from_mp3(src)
    else:
        raise InvalidExtensionError("{} has {} extension".format(src, extension))

    filename = get_file_name(src)
    audio_file += units
    if units < 0:
        modified_file = dst + "\\{}_minus_{}.{}".format(filename, -units, extension)
    else:
        modified_file = dst + "\\{}_plus_{}.{}".format(filename, units, extension)

    audio_file.export(modified_file, extension)

    # code=0
    return True





# TESTING
if __name__ == '__main__':
    p = "path to wav"
    # save_text_to_speech("hello my name is andrew", p, __print=True)
    # sleep(1)
    t = speech_to_text_from_file(p)
    print(t)



"""
    core/symbols.py

    static symbols that can be
    imported to improve de design
    of a program

    author: @alexzander
"""


# core package (pip install python-core)
from core.aesthetics import *


# ═
equal_symbol = chr(9552)
# ═┳═
latteral_connection_symbol = chr(9552) + chr(9523) + chr(9552)
# ┗═
down_right_connection_symbol = chr(9495) + chr(9552)
# ⋘
triple_left_shift = chr(8920)
# ⋙
triple_right_shift = chr(8921)


# [
left_bracket_cyan = ConsoleColored("[", "cyan", bold=1)
# ]
right_bracket_cyan = ConsoleColored("]", "cyan", bold=1)


# -
dash_orange = ConsoleColored("-", "red", bold=1)


# <
left_arrow_blue = ConsoleColored("<", "blue", bold=1)
# <
left_arrow_yellow = ConsoleColored("<", "yellow", bold=1)
# >
right_arrow_blue = ConsoleColored(">", "blue", bold=1)
# >
right_arrow_yellow = ConsoleColored(">", "yellow", bold=1)

# >>>
left_arrow_3_green_bold = green_bold(">>>")




import os
from string import Template
from core.__path import *


function_template = """
def {filename}_sound():
    play_commands(remote_sounds_json, "{filename}")

"""

def AutoGenerate(py_file: str, database_folder: dict):
    if not is_file(py_file):
        raise NotAFileError
    if not py_file.endswith(".py") and not py_file.endswith(".pyw"):
        raise IncorrectExtensionError
    
    with open(py_file, "a+", encoding="utf-8") as f:
        
        for audio_file in os.listdir(database_folder):
            filename = audio_file.split(".")[0] # get rid of extension
            if "-" in filename:
                filename = filename.replace("-", "_")
            f.write(function_template.format(filename=filename))



"""
    core/romania.py

    helpful with extracting
    birth date from
    social security numbers

    also contains every district
    from this country

    good for romanian developers or
    people who live in romania

    author: @alexzander
"""


romanian_diacritics = {
    "a": ["ă", "â"],
    "A": ["Ă", "Â"],
    "i": "î",
    "I": "Î",
    "s": "ş",
    "S": "Ş",
    "t": "ţ",
    "T": "Ţ"
}


romanian_districts = [
    {
        "name": "Alba",
        "code": "01",
    },
    {
        "name": "Arad",
        "code": "02"
    },
    {
        "name": "Arge\u0219",
        "code": "03"
    },
    {
        "name": "Bac\u0103u",
        "code": "04"
    },
    {
        "name": "Bihor",
        "code": "05"
    },
    {
        "name": "Bistri\u021ba-N\u0103s\u0103ud",
        "code": "06"
    },
    {
        "name": "Boto\u0219ani",
        "code": "07"
    },
    {
        "name": "Bra\u0219ov",
        "code": "08"
    },
    {
        "name": "Br\u0103ila",
        "code": "09"
    },
    {
        "name": "Buz\u0103u",
        "code": "10"
    },
    {
        "name": "Cara\u0219-Severin",
        "code": "11"
    },
    {
        "name": "Cluj",
        "code": "12"
    },
    {
        "name": "Constan\u021ba",
        "code": "13"
    },
    {
        "name": "Covasna",
        "code": "14"
    },
    {
        "name": "D\u00e2mbovi\u021ba",
        "code": "15"
    },
    {
        "name": "Dolj",
        "code": "16"
    },
    {
        "name": "Gala\u021bi",
        "code": "17"
    },
    {
        "name": "Gorj",
        "code": "18"
    },
    {
        "name": "Harghita",
        "code": "19"
    },
    {
        "name": "Hunedoara",
        "code": "20"
    },
    {
        "name": "Ialomi\u021ba",
        "code": "21"
    },
    {
        "name": "Ia\u0219i",
        "code": "22"
    },
    {
        "name": "Ilfov",
        "code": "23"
    },
    {
        "name": "Maramure\u0219",
        "code": "24"
    },
    {
        "name": "Mehedin\u021bi",
        "code": "25"
    },
    {
        "name": "Mure\u0219",
        "code": "26"
    },
    {
        "name": "Neam\u021b",
        "code": "27"
    },
    {
        "name": "Olt",
        "code": "28"
    },
    {
        "name": "Prahova",
        "code": "29"
    },
    {
        "name": "Satu Mare",
        "code": "30"
    },
    {
        "name": "S\u0103laj",
        "code": "31"
    },
    {
        "name": "Sibiu",
        "code": "32"
    },
    {
        "name": "Suceava",
        "code": "33"
    },
    {
        "name": "Teleorman",
        "code": "34"
    },
    {
        "name": "Timi\u0219",
        "code": "35"
    },
    {
        "name": "Tulcea",
        "code": "36"
    },
    {
        "name": "Vaslui",
        "code": "37"
    },
    {
        "name": "V\u00e2lcea",
        "code": "38"
    },
    {
        "name": "Vrancea",
        "code": "39"
    },
    {
        "name": "Bucure\u0219ti",
        "code": "40"
    },
    {
        "name": "Bucure\u0219ti - Sector 1",
        "code": "41"
    },
    {
        "name": "Bucure\u0219ti - Sector 2",
        "code": "42"
    },
    {
        "name": "Bucure\u0219ti - Sector 3",
        "code": "43"
    },
    {
        "name": "Bucure\u0219ti - Sector 4",
        "code": "44"
    },
    {
        "name": "Bucure\u0219ti - Sector 5",
        "code": "45"
    },
    {
        "name": "Bucure\u0219ti - Sector 6",
        "code": "46"
    },
    {
        "name": "C\u0103l\u0103ra\u0219i",
        "code": "51"
    },
    {
        "name": "Giurgiu",
        "code": "52"
    }
]


def birth_date_from_cnp(code):
    # 1800 -> 1899
    if code[0] == "3" or code[0] == "4":
        an = 1800
        if int(code[1]) == 0 and int(code[2]) != 0:
            an += int(code[2])
        elif int(code[1]) != 0 and int(code[2]) != 0:
            an += int(code[1:3])
        luna = int(code[3:5]) if int(code[3]) != 0 else int(code[4])
        zi = int(code[5:7]) if int(code[5]) != 0 else int(code[6])
        return an, luna, zi

    # 1900 -> 1999
    elif code[0] == "1" or code[0] == "2":
        an = 1900
        if int(code[1]) == 0 and int(code[2]) != 0:
            an += int(code[2])
        elif int(code[1]) != 0 and int(code[2]) != 0:
            an += int(code[1:3])
        luna = int(code[3:5]) if int(code[3]) != 0 else int(code[4])
        zi = int(code[5:7]) if int(code[5]) != 0 else int(code[6])
        return an, luna, zi
    # 2000 -> 2099
    elif code[0] == "5" or code[0] == "6":
        an = 2000
        if int(code[1]) == 0 and int(code[2]) != 0:
            an += int(code[2])
        elif int(code[1]) != 0 and int(code[2]) != 0:
            an += int(code[1:3])
        luna = int(code[3:5]) if int(code[3]) != 0 else int(code[4])
        zi = int(code[5:7]) if int(code[5]) != 0 else int(code[6])
        return an, luna, zi
    return None


# TESTING
if __name__ == '__main__':
    pass




class Node:
    def __init__(self, data=None):
        self.data = data
        self.next = None


    def __str__(self):
        content = f"Node({self.data})"
        return content


class LinkedList:
    def __init__(self):
        self.head = Node()


    def append(self, data):
        current_node = self.head
        while current_node.next:
            current_node = current_node.next
        current_node.next = Node(data)


    def __len__(self):
        current_node = self.head
        length = 0
        while current_node.next:
            length += 1
            current_node = current_node.next
        return length


    def __str__(self):
        content = "["
        current_node = self.head
        while current_node.next:
            current_node = current_node.next
            content += str(current_node.data) + "->"
        content = content[:len(content) - 2] + "]"
        if content:
            return content
        return "there is just an empty head"


    def __getitem__(self, index: int):
        if not (0 <= index <= len(self) - 1):
            raise IndexError("linked list out of range in getting item")
        current_index = 0
        current_node = self.head
        while current_node.next:
            current_node = current_node.next
            if current_index == index:
                return current_node.data
            current_index += 1


    # concatenating two linked lists
    def __add__(self, other_head):
        current_node = self.head
        while current_node.next:
            current_node = current_node.next
        current_node.next = other_head.head.next
        return self


    # deletes the item and returns it
    def __delitem__(self, index):
        """ how to use: 'del linked_list[eBookIndex]' """
        if not (0 <= index <= len(self) - 1):
            raise IndexError("linked list out of range in deleting item")
        current_index = 0
        current_node = self.head
        while current_node.next:
            last_node = current_node
            current_node = current_node.next
            if current_index == index:
                last_node.next = current_node.next
                return current_node.data
            current_index += 1


    def delete_item(self, index):
        del self[index]


    def insert_item(self, data, index):
        if not (0 <= index <= len(self) - 1):
            raise IndexError("linked list out of range on inserting an item")
        if not data:
            raise ValueError("None value on inserting an item")

        current_index = 0
        current_node = self.head
        while current_node.next:
            last_node = current_node
            current_node = current_node.next
            if current_index == index:
                new_node = Node(data)
                last_node.next = new_node
                new_node.next = current_node
                return
            current_index += 1


    def insert_node(self, node: Node, index):
        if not (0 <= index <= len(self) - 1):
            raise IndexError("linked list out of range on inserting an item")
        if not node.data:
            raise ValueError("None value on inserting an item")

        current_index = 0
        current_node = self.head
        while current_node.next:
            last_node = current_node
            current_node = current_node.next
            if current_index == index:
                last_node.next = node
                node.next = current_node
                return
            current_index += 1


    # returns the old item
    def overwrite_item(self, index, data):
        if not (0 <= index <= len(self) - 1):
            raise IndexError("linked list out of range on inserting an item")
        if not data:
            raise ValueError("None value on inserting an item")

        current_index = 0
        current_node = self.head
        while current_node.next:
            current_node = current_node.next
            if current_index == index:
                old_item = current_node.data
                current_node.data = data
                return old_item
            current_index += 1


# TESTING
if __name__ == '__main__':
    ll = LinkedList()
    ll.append(123)
    ll.append(676)
    ll.append(23)
    ll.append(23123)
    ll.append(2323)
    ll.append(2320000003)
    print(ll)
    for index in range(len(ll)):
        print(index, "=>", end='')
    print()
    ll.insert_item(74, 1)
    print(ll)

    item = Node(1)
    print(item)

    ll.insert_node(item, 4)
    print(ll)

    ll.overwrite_item(1, 7474)
    print(ll)



"""
    core/gui/configs.py

    configurations 4 tkinter objects
    helpful in dev of guis

    author: @alexzander
"""


# python
from tkinter import *


# consolas
consolas_10_bold = ('Consolas', 10, 'bold')
consolas_20_bold = ('Consolas', 20, 'bold')
consolas_30_bold = ('Consolas', 30, 'bold')
consolas_40_bold = ('Consolas', 40, 'bold')
consolas_50_bold = ('Consolas', 50, 'bold')
consolas_60_bold = ('Consolas', 60, 'bold')


# cascadia code
cascadia_code_10 = ('Cascadia Code', 10)
cascadia_code_10_bold = ('Cascadia Code', 10, "bold")

cascadia_code_20 = ('Cascadia Code', 20)
cascadia_code_20_bold = ('Cascadia Code', 20, 'bold')

cascadia_code_30 = ('Cascadia Code', 30)
cascadia_code_30_bold = ('Cascadia Code', 30, 'bold')

cascadia_code_40 = ('Cascadia Code', 40)
cascadia_code_40_bold = ('Cascadia Code', 40, 'bold')

cascadia_code_50 = ('Cascadia Code', 50)
cascadia_code_50_bold = ('Cascadia Code', 50, 'bold')


# abadi
abadi_45_bold = ("Abadi", 45, 'bold')


# colors
red = "red"
white = "white"
yellow = "yellow"
blue = "blue"
black = "black"
lightgreen = 'lightgreen'
gray32 = "gray32"
light_golden_rod = "light goldenrod"
gold="gold"
state_blue = "state blue"
indian_red = "indian red"
green = "green"
green2 = "green2"
thistle2 = "thistle2"
gray63 = "gray63"
seashell2 = "seashell2"
gray10 = "gray10"
tomato2 = "tomato2"
pale_green = "pale green"
gold2 = "gold2"
spring_green = "spring green"
green_yellow = "green yellow"
forest_green = "forest green"
brown4 = "brown4"
dark_green = "dark green"
gold3 = "gold3"
DodgerBlue2 = "DodgerBlue2"
gray57 = "gray57"


def config_gui_obj(gui_obj, font: tuple, back=seashell2, fore=gray10,
                   text="", width=None, height=None):
    if height:
        gui_obj.config(height=height)
    if width:
        gui_obj.config(width=width)
    gui_obj.config(font=font, back=back, fore=fore, text=text)



"""
    core/backtracking.py

    this is the old style of C++
    in using backtracking implemented
    in python3

    author: @alexzander
"""


class Permutations:
    def __init__(self, dimension=3):
        self.dimension = dimension # old 'N'
        self.position = 0 # old 'K'
        self.solution = [0 for _ in range(self.dimension)] # old 'stiva'
        # these are old elements from c++

    def Init(self):
        self.solution[self.position] = 0

    def HasSuccesor(self):
        if self.solution[self.position] < self.dimension:
            self.solution[self.position] += 1
            return True
        else:
            return False

    def ValidSolution(self):
        for index in range(0, self.position):
            if self.solution[index] == self.solution[self.position]:
                return False
        return True

    def Solution(self):
        return self.dimension - 1 == self.position

    def Permutations(self):
        self.__permutations()
        self.__init__()

    def __permutations(self):
        self.position = 0
        self.Init()
        while self.position > -1:
            while True:
                succesor = self.HasSuccesor()
                if not succesor or self.ValidSolution():
                    break
            if succesor:
                if self.Solution():
                    print(self.solution)
                else:
                    self.position += 1
                    self.Init()
            else:
                self.position -= 1


class Arrangements:
    def __init__(self, dimension, solution_size):
        self.dimension = dimension  # old 'N'
        self.position = 0  # old 'K'
        self.solution = [0 for _ in range(self.dimension)]  # old 'stiva'
        # these are old elements from c++
        # we have in extra a different size for every solution
        self.solution_size = solution_size

    def Init(self):
        self.solution[self.position] = 0

    def HasSuccesor(self):
        if self.solution[self.position] < self.dimension:
            self.solution[self.position] += 1
            return True
        else:
            return False

    def ValidSolution(self):
        for index in range(self.position):
            if self.solution[index] == self.solution[self.position]:
                return False
        return True

    def Solution(self):
        return self.solution_size - 1 == self.position

    def Arragements(self):
        self.__arrangements()
        self.__init__(self.dimension, self.solution_size)

    def __arrangements(self):
        self.position = 0
        self.Init()
        while self.position > -1:
            while True:
                succesor = self.HasSuccesor()
                if not succesor or self.ValidSolution():
                    break
            if succesor:
                if self.Solution():
                    print(self.solution[:self.solution_size])
                else:
                    self.position += 1
                    self.Init()
            else:
                self.position -= 1


class Combinations:
    def __init__(self, dimension, solution_size):
        self.dimension = dimension  # old 'N'
        self.position = 0  # old 'K'
        self.solution = [0 for _ in range(self.dimension)]  # old 'stiva'
        # these are old elements from c++
        # we have in extra a different size for every solution
        self.solution_size = solution_size

    def Init(self):
        if self.position == 0:
            self.solution[0] = 0
        else:
            self.solution[self.position] = self.solution[self.position - 1]

    def HasSuccesor(self):
        if self.solution[self.position] < self.dimension:
            self.solution[self.position] += 1
            return True
        return False

    def ValidSolution(self):
        for index in range(self.position):
            if self.solution[index] == self.solution[self.position]:
                return False
        return True

    def Solution(self):
        return self.solution_size - 1 == self.position

    def Combinations(self):
        self.__combinations()
        self.__init__(self.dimension, self.solution_size)

    def __combinations(self):
        self.position = 0
        self.Init()
        while self.position > -1:
            while True:
                succesor = self.HasSuccesor()
                if not succesor or self.ValidSolution():
                    break
            if succesor:
                if self.Solution():
                    print(self.solution[:self.solution_size])
                else:
                    self.position += 1
                    self.Init()
            else:
                self.position -= 1


class NumberSumPartitions:
    def __init__(self, number):
        self.number = number
        self.dimension = number
        self.position = 0
        self.solution = [0 for _ in range(number)]
        self.index = 0
        self.solutions = []

    def Init(self):
        self.solution[self.position] = 0

    def HasSuccesor(self):
        if self.solution[self.position] < self.dimension:
            self.solution[self.position] += 1
            return True
        return False

    def Solution(self):
        partition_sum = 0
        for index in range(len(self.solution)):
            partition_sum += self.solution[index]
            if partition_sum == self.number:
                self.index = index
                solut = sorted(self.solution[:index + 1])
                if solut not in self.solutions:
                    print(solut)
                    self.solutions.append(solut)
                return True
        return False

    def Partitions(self):
        self.__partitions()
        self.__init__(self.number)

    def __partitions(self):
        while self.position > -1:
            succesor = self.HasSuccesor()
            if succesor:
                if self.Solution():
                    # automatically prints solution
                    pass
                else:
                    if self.position < len(self.solution) - 1:
                        self.position += 1
                        self.Init()
            else:
                self.position -= 1


class MadameBacktracking:
    def __init__(self, dimension=3):
        self.dimension = dimension
        self.position = 0
        self.solution = [0 for _ in range(self.dimension)]

    def Init(self):
        self.solution[self.position] = 0

    def HasSuccesor(self):
        if self.solution[self.position] < self.dimension:
            self.solution[self.position] += 1
            return True
        return False

    def ValidSolution(self):
        for index in range(self.position):
            if self.solution[index] == self.solution[self.position] \
                or abs(self.solution[self.position] - self.solution[index]) == self.position - index:
                return False
        return True

    def Solution(self):
        return self.dimension - 1 == self.position

    def MadameBacktracking(self):
        self.__madame_backtracking()
        self.__init__(self.dimension)

    def __madame_backtracking(self):
        self.position = 0
        self.Init()
        while self.position > -1:
            while True:
                succesor = self.HasSuccesor()
                if not succesor or self.ValidSolution():
                    break
            if succesor:
                if self.Solution():
                    solution = list(map(lambda item: item - 1, self.solution))
                    mat_visualize = [[0 for _ in range(self.dimension)] for _ in range(self.dimension)]
                    for x, y in zip(range(self.dimension), solution):
                        mat_visualize[x][y] = 1
                    for line in mat_visualize:
                        print(line)
                    print()
                else:
                    self.position += 1
                    self.Init()
            else:
                self.position -= 1


class PermutationsRecursive:
    def __init__(self, dimension):
        self.dimension = dimension
        self.position = 0
        self.solution = [0 for _ in range(self.dimension)]
        self.solutions = []
        self.__permutations(self.position)

    def Continue(self, position):
        for index in range(position):
            if self.solution[index] == self.solution[position]:
                return False
        return True

    def __permutations(self, position):
        if position == self.dimension:
            self.solutions.append(list(self.solution))
        else:
            for index in range(1, self.dimension + 1):
                self.solution[position] = index
                if self.Continue(position):
                    self.__permutations(position + 1)

    def __iter__(self):
        if self.solutions:
            return iter(self.solutions)
        return []

    def __str__(self):
        if self.solutions:
            for permutation in self:
                print(permutation)
        else:
            print()
        return ""

    def __set__(self, __object):
        if self.solutions:
            from copy import deepcopy
            __object = deepcopy(self.solutions)
        else:
            __object = []



# TESTING
if __name__ == '__main__':
    Permutations(6).Permutations()



"""
    core/session.py

    wrapped around bs4

    author: @alexzander
"""


# 3rd party
from bs4 import BeautifulSoup # pip install bs4
from requests_html import HTMLSession # pip install requrests_html

# core package ( pip install python-core )
from core.aesthetics import ConsoleColored


def get_soup(url, __print=False):
    fancy_url = ConsoleColored(url, "yellow", bold=1, underlined=1)
    if __print:
        print("HTTP GET request to URL: {} ...".format(fancy_url))

    session = HTMLSession()
    response = session.get(url)
    response.html.render(sleep=0.0001)
    if __print:
        print("creating soup...")
    soup = BeautifulSoup(response.html.html, "html.parser")
    if __print:
        print(ConsoleColored("soup created successfully.", "yellow", bold=1))
    return soup



"""
    core/__datetime.py

    powerful time, date and datetime module
    useful in development of programs that
    work with the concept of time

    upgraded version of time.py and datetime.py

    author: @alexzader
"""


# python
import time
import calendar
from datetime import datetime

# core package ( pip install python-core )
from core.__numbers import *
from core.aesthetics import *


date_format = "%d.%m.%Y"
time_format = "%H:%M:%S"

datetime_format = "[ {} ] - [ {} ]".format(date_format, time_format)
timedate_format = "[ {} ] - [ {} ]".format(time_format, date_format)


def is_valid_date(date_str: str):
    try:
        datetime.strptime(date_str, date_format)
        return True
    except:
        return False


def is_valid_time(time_str: str):
    try:
        datetime.strptime(time_str, time_format)
        return True
    except:
        return False


def is_valid_datetime(datetime_str: str):
    try:
        datetime.strptime(datetime_str, datetime_format)
        return True
    except:
        return False


def get_current_date(__format=date_format):
    return datetime.now().strftime(__format)


def get_current_time(__format=time_format):
    return datetime.now().strftime(__format)


def get_current_datetime(__format=datetime_format):
    return datetime.now().strftime(__format)


def get_current_timedate(__format=timedate_format):
    return datetime.now().strftime(__format)


def get_execution_time(__function, *params):
    before = time()
    result = __function(*params)
    if result != None:
        print(result)

    duration = time() - before
    duration = fixed_set_precision_str(duration, 2)
    return duration


def print_execution_time(__function, *params):
    print("execution time: [ {} second(s) ]".format(yellow_bold(get_execution_time(__function, *params))))



"""
    core/__winapi.py
    useful windows api file

    author:@alexzander
"""


# 3rd party
from win10toast import ToastNotifier # pip install win10toast


def windows_notification(title, message, duration, icon, threaded=True):
    ToastNotifier().show_toast(title=title, msg=message, duration=duration, icon_path=icon, threaded=threaded)



# TESTING
if __name__ == '__main__':
    pass



"""
    system.py

    wheather module wrapped around
    pyowm ( open wheather map api)

    author: @alexzander
"""


# 3rd party
from pyowm import OWM

# core package (pip install python-core)
from core.system import *


class Wheather(OWM):
    def __init__(self, api_key):
        self.api_key = api_key
        super().__init__(self.api_key)
        self.manager = self.weather_manager()


    def get_temperature(self, location, unit_measurement="celsius"):
        w = self.manager.weather_at_place(location).weather
        return w.temperature(unit_measurement)["temp"]



# TESTING
if __name__ == '__main__':
    # usage
    api_key = "your api key from open weather map website"
    w = Wheather(api_key)
    t = w.get_temperature("Aalbord")
    print(t)


import setuptools

with open("README.md", "r", encoding="utf-8") as f:
    long_description = f.read()

setuptools.setup(
    name="python-core", # your package name
    version="0.0.3", # your package version
    author="alexzander", # package author
    author_email="just.python.mail.test@gmail.com", # package author email
    description="", # package description
    long_description=long_description, # package long description
    long_description_content_type="text/markdown",
    url="",
    include_package_data=True,
    install_requires=[ # all installed 3rd party modules
        "aiohttp",
		"altgraph",
		"appdirs",
		"argon2-cffi",
		"args",
		"asgiref",
		"astroid",
		"async-timeout",
		"attrs",
		"autopep8",
		"backcall",
		"beautifulsoup4",
		"bitly-api",
		"bleach",
		"bs4",
		"cachetools",
		"cairocffi",
		"CairoSVG",
		"certifi",
		"cffi",
		"chardet",
		"click",
		"clint",
		"colorama",
		"colored-traceback",
		"comtypes",
		"cssselect",
		"cssselect2",
		"datedelta",
		"decorator",
		"defusedxml",
		"Django",
		"docutils",
		"entrypoints",
		"fake-useragent",
		"ffmpeg-python",
		"future",
		"geojson",
		"googletrans",
		"gTTS",
		"h11",
		"h2",
		"hpack",
		"hstspreload",
		"httpcore",
		"httpx",
		"hyperframe",
		"idna",
		"imageio",
		"imageio-ffmpeg",
		"importlib-metadata",
		"InstagramAPI",
		"ipinfo",
		"ipykernel",
		"ipython",
		"ipython-genutils",
		"ipywidgets",
		"isort",
		"jedi",
		"Jinja2",
		"jsonschema",
		"jupyter-client",
		"jupyter-console",
		"jupyter-core",
		"keyboard",
		"keyring",
		"Kivy",
		"kivy-deps.angle",
		"kivy-deps.glew",
		"kivy-deps.gstreamer",
		"kivy-deps.sdl2",
		"Kivy-examples",
		"Kivy-Garden",
		"lazy-object-proxy",
		"lxml",
		"MarkupSafe",
		"mccabe",
		"mistune",
		"MouseInfo",
		"moviepy",
		"multidict",
		"mysqlclient",
		"nbconvert",
		"nbformat",
		"notebook",
		"numpy",
		"opencv-python",
		"packaging",
		"pandocfilters",
		"parse",
		"parso",
		"pdfminer.six",
		"pdfplumber",
		"pefile",
		"pickleshare",
		"Pillow",
		"pkginfo",
		"playsound",
		"proglog",
		"prometheus-client",
		"prompt-toolkit",
		"psutil",
		"PyAudio",
		"PyAutoGUI",
		"pycodestyle",
		"pycparser",
		"pycryptodome",
		"pydub",
		"pyee",
		"pygal",
		"pygame",
		"PyGetWindow",
		"Pygments",
		"pygubu",
		"pygubu-designer",
		"pyinstaller",
		"pyinstaller-hooks-contrib",
		"pylint",
		"PyMsgBox",
		"pynput",
		"pyowm",
		"pyparsing",
		"PyPDF2",
		"pyperclip",
		"pypiwin32",
		"pyppeteer",
		"PyQt5",
		"PyQt5-sip",
		"pyquery",
		"PyRect",
		"pyrsistent",
		"PyScreeze",
		"PySocks",
		"pytesseract",
		"python-core",
		"python-dateutil",
		"pyttsx3",
		"PyTweening",
		"pytz",
		"pywin32",
		"pywin32-ctypes",
		"pywinpty",
		"pyzbar",
		"pyzmq",
		"qtconsole",
		"QtPy",
		"readme-renderer",
		"requests",
		"requests-html",
		"requests-toolbelt",
		"rfc3986",
		"scipy",
		"selenium",
		"Send2Trash",
		"six",
		"sniffio",
		"sortedcontainers",
		"SoundCard",
		"sounddevice",
		"soupsieve",
		"SpeechRecognition",
		"spotipy",
		"sqlparse",
		"terminado",
		"testpath",
		"tinycss2",
		"toml",
		"tornado",
		"tqdm",
		"traitlets",
		"twine",
		"typed-ast",
		"typing-extensions",
		"urllib3",
		"vboxapi",
		"w3lib",
		"Wand",
		"wcwidth",
		"webencodings",
		"websockets",
		"widgetsnbextension",
		"win10toast",
		"windows-curses",
		"wrapt",
		"yarl",
		"zipp"
    ],
    platforms="any",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.7',
)



"""
    core/development.py

    useful module in development
    of executable python3 applications

    author: @alexzander
"""


# core package ( pip install python-core )
from core.aesthetics import *


def IntroductionMessage(message: str):
    if message == "":
        raise ValueError

    message = "\n" + message
    print_red_bold(message)
    print_yellow_bold("{}\n".format("=" * (len(message) - 1)))



"""
    This module contains an upgraded version of the 
        primitive built-in list from python.
    This module also contains a custom-built error 
        used only for the class List.
    
    Current state: in development.
"""


# python
import random
from sys import getsizeof
# core package
from core.__collections.__list.exceptions import *


class List(list):
    """ This is the implementation of a better python list. 
        Made in python ofc. """
        
    def __init__(self, *objects):
        self.container = []
        if len(objects) == 1:
            collection = objects[0] 
            if type(collection) == list:
                self.container = list(collection)
            elif type(collection) == List:
                self.container = [x for x in collection.container]
            elif type(collection) == str:
                self.container = list(collection)
            elif type(collection) == int:
                while collection != 0:
                    self.container.append(collection % 10)
                    collection //= 10
                self.container = self.container[::-1]
        elif len(objects) > 1:
            for obj in objects:
                self.container.append(obj)
                
        
        # keeping dimension in a variable 
        # for not calling __len__ so many times
        # its inefficent...
        self.dimension = len(self.container)


    def append(self, argument, *arguments, **keywordArguments):
        """
            Adds an element to the end of the list, is 
                compulsory to pass at least one argument to 
                the append List method.
            Adds multiple elements to the end of the list, 
                if existent.
            With python built-in list you can't do that, fools.
            Adds multiple keyword arguments to the end 
                of the list, if existent.
            Complexity: Dependent of the length of the 
                *arguments and **keywordArguments.
        """
        counter = 0
        self.container.append(argument)
        for obj in arguments:
            self.container.append(obj)
            counter += 1
        if keywordArguments != {}:
            dictionary = {}
            for key, arg in keywordArguments.items():
                dictionary.update({key: arg})
            self.container.append(dictionary)
            counter += 1    
        self.dimension += counter


    def append_front(self, argument, *arguments, **keywordArguments):
        """
            Adds an element to the beginning of the list, 
                is compulsory to pass at least one argument to the append_front List method.
            Adds multiple elements to the beginning of the 
                list, if existent.
            With python built-in list you can't do that, fools.
            Adds multiple keyword arguments to the 
                beginning of the list, if existent.
            Complexity: Dependent of the length of the 
                *arguments and **keywordArguments.
        """
        self.container.insert(0, argument)
        if arguments != []:
            counter = 1
            for obj in arguments:
                self.container.insert(counter, obj)
                counter += 1
        times = counter
        if keywordArguments != {}:
            dictionary = {}
            for key, argument in keywordArguments.items():
                dictionary.update({key: argument})
            self.container.insert(counter, dictionary)
            times += 1
        self.dimension += times


    def extend(self, iterable):
        if type(iterable) == List:
            self.container.extend(iterable.container)
        else:
            self.container.extend(iterable)


    def insert(self, index: int, *arguments, **keywordArguments):
        """ Inserting elements on the specified @param index.
            if index == len(self.container):
                then we append the element
            This function is capable of inserting an infinite number
            of arguments on the specified position.
            And also keyword arguments.
        """
        if index == len(self.container):
            for argument in arguments:
                self.container.append(argument)
            if keywordArguments != {}:
                dictionary = {}
                for key, argument in keywordArguments.items():
                    dictionary.update({key: argument})
                self.container.append(dictionary)
        else:
            # we cant insert in an empty list on 
            # a bigger index than zero
            if self.container == [] and index > 0:
                raise IndexError
            counter = 0
            for argument in arguments:
                self.container.insert(index + counter, argument)
                counter += 1
            if keywordArguments != {}:
                dictionary = {}
                for key, argument in keywordArguments.items():
                    dictionary.update({key: argument})
                self.container.insert(counter, dictionary)


    def remove(self, item):
        if self.container == []:
            raise EmptyListError
        self.container.remove(item)

 
    def pop(self, index=None):
        if self.container == []:
            raise EmptyListError
        
        if index is None:
            index = len(self.container) - 1
        if not (0 <= index < len(self.container)):
            raise IndexError
        
        self.dimension -= 1
        return self.container.pop(index=index)


    def clear(self):
        self.container.clear()
        self.dimension = 0


    def index(self, item, start=None, stop=None):
        """ Just the old list.index method with the same 
            implementation. """
        if start is None and stop is None:
            return self.container.index(item)
        elif start is not None and stop is None:
            if type(start) != int:
                raise TypeError("@param start is not int.")
            if 0 <= start < len(self.container):
                return self.container.index(item, start=start)
            else:
                raise IndexError
        elif start is not None and stop is not None:
            if type(start) != int or type(stop) != int:
                raise TypeError("@params start and stop are not int")
            if 0 <= start < stop < len(self.container):
                return self.container.index(item, start, stop)
            else:
                raise IndexError
        else:
            raise TypeError


    def count(self, item):
        if self.container == []:
            raise EmptyListError
        return self.container.count(item)


    def sort(self, key=None, reverse=False):
        if self.container == []:
            raise EmptyListError
        self.container.sort(key=key, reverse=reverse)


    def reverse(self):
        if self.container == []:
            raise EmptyListError
        self.container.reverse()


    def copy(self):
        return self.container.copy()


    def __setitem__(self, *args):
        if len(args) == 1:
            instance = args[0]
            if type(instance) == list:
                self.container = instance
            elif type(instance) == List:
                self.container = instance.container
            else:
                raise TypeError
            
        elif len(args) == 2:
            if type(args[0]) == slice:
                if self.container == []:
                    raise EmptyListError
                
                sl = args[0]
                start, stop, step = sl.start, sl.stop, sl.step
                temporary = args[1]
                if type(temporary) == List:
                    temporary = temporary.container
                elif type(temporary) != list and type(temporary) != tuple and type(temporary) != dict and type(temporary) != set:
                    self.container.append(temporary)
                    return
                
                if stop is None:
                    self.container[start::step] = temporary
                else:
                    self.container[start:stop - 1:step] = temporary
            elif type(args[0]) == int:
                if not (0 <= args[0] < len(self.container)):
                    raise IndexError
                self.container[args[0]] = args[1]


    def __delitem__(self, index):
        if type(index) == int:
            del self.container[index]
        elif type(index) == slice:
            start, stop, step = index.start, index.stop, index.step
            counter = 0
            if stop is None:
                for _ in range(start, len(self.container), step):
                    counter += 1
                del self.container[start::step]
            else:
                for _ in range(start, stop, step):
                    counter += 1
                del self.container[start:stop - 1:step]
            self.dimension -= counter
        else:
            raise TypeError


    def last_item(self):
        """ Returns the last element of the list.
            If its a stack its top of the list. 
            If the list is empty raises EmptyListError"""
        if self.container == []:
            raise EmptyListError
        
        return self.container[len(self.container) - 1]


    def __str__(self):
        return self.container.__str__()


    def __repr__(self):
        return self.container.__repr__()


    # ==
    def __eq__(self, instance):
        if type(instance) == list:
            return self.container == instance
        elif type(instance) == List:
            return self.container == instance.container
        else:
            raise TypeError


    # greater than >
    def __gt__(self, instance):
        if type(instance) == list:
            return self.container > instance
        elif type(instance) == List:
            return self.container > instance.container
        else:
            raise TypeError


    # less than <
    def __lt__(self, instance):
        if type(instance) == list:
            return self.container < instance
        elif type(instance) == List:
            return self.container < instance.container
        else:
            raise TypeError


    # less equal <=
    def __le__(self, instance):
        if type(instance) == list:
            return self.container <= instance
        elif type(instance) == List:
            return self.container <= instance.container
        else:
            raise TypeError


    # greater equal >=
    def __ge__(self, instance):
        if type(instance) == list:
            return self.container >= instance
        elif type(instance) == List:
            return self.container >= instance.container
        else:
            raise TypeError


    # not equal (not ==)
    def __ne__(self, instance):
        if type(instance) == list:
            return self.container != instance
        elif type(instance) == List:
            return self.container != instance.container
        else:
            raise TypeError


    def __del__(self):
        pass


    def format(self, *arguments):
        """ 
            Formating on list. 
            This mechanism is applied only empty dictionaries present in list.
        """
        empty_dicts = []
        for index in range(len(self.container)):
            if self.container[index] == {}:
                empty_dicts.append(index)
        for index, argument in zip(empty_dicts, arguments):
            self.container[index] = argument


    # calling str.format(List)
    def __format__(self, format_spec):
        return "<List self.container={}>".format(self.container)


    def __sizeof__(self):
        return object.__sizeof__(self) + \
            sum(getsizeof(value) for value in self.__dict__.values()) + \
            sum(getsizeof(item) for item in self.container)


    def __getitem__(self, index: int):
        if self.container != []:
            if type(index) == slice:
                start, stop, step = (index.start, index.stop, index.step)
                # print(start, stop, step)
                # print(type(start), type(stop), type(step))
                # similar implementation of range class
                # we did that before
                
                #TODO fix this
                # if (type(start) != int and type(start) != None) or (type(stop) != int and type(stop) != None) or (type(step) != int and type(step) != None):
                #     raise TypeError("Neither int or None")
                
                if stop is not None:
                    return self.container[start:stop - 1:step]
                return self.container[start::step]
                
            elif type(index) == int:
                if 0 <= index < len(self.container):
                    return self.container[index]
                else:
                    raise IndexError                
        else:
            raise EmptyListError("You cannot get an item of an empty list.")


    def __len__(self):
        return self.container.__len__()


    def __add__(self, instance):
        """ Returns the concatenation of both self and instance 
            as a new List object. 
            Example:
                l1 = [1, 2, 3, 4]
                l2 = [5, 6, 7, 8]
                l1 + l2 == [1, 2, 3, 4, 5, 6, 7, 8]
        """
        iterable = instance
        if type(instance) == List:
            iterable = instance.container
        elif type(instance) == tuple:
            pass
        elif type(instance) != list:
            raise TypeError
            
        return List(self.container + iterable)


    def __sub__(self, instance):
        """ Returns the elements which are existent in 
            @param self and NOT existent in @param instance. 
            Example:
                self=[4, 5, 6, 7]
                instance=[4, 5, 8, 9]
                self - instance == [6, 7]
        """
        if self.container == []:
            raise EmptyListError
            
        iterable = instance
        if type(instance) == List:
            iterable = instance.container
        elif type(instance) == tuple:
            pass
        elif type(instance) != list:
            raise TypeError
        
        return List([x for x in self.container if x not in iterable])


    def __mul__(self, instance):
        """ Application of * on List.
            if instance is integer:
                then we return multiplication of the
                self.container times integer
                
                Example:
                [1, 2] * 4 == [1, 2, 1, 2, 1, 2, 1, 2]
                
            if instance is list or List:
                then we return parallel multiplication of every
                element from self.container and instance
                
                Example:
                [3, 4] * [4, 5] == [12, 20]
                
                if their lengths are not equal
                    zip function is taking care of 
                    
                    Example:
                    [1, 2, 3, 4] * [2, 2] == [2, 4]
                    [5, 5] * [6, 7, 7, 7] == [30, 35]
                    
            if we want to multiply the strings the list
                
                Example:
                ["s", "S", "a"] * 3 == ["sss", "SSS", "aaa"]
        """
        if self.container == []:
            raise EmptyListError
        
        if type(instance) == int:
            return List(self.container * instance)
        else:
            iterable = instance
            if type(instance) == List:
                iterable = instance.container
            elif type(iterable) == tuple:
                pass
            elif type(instance) != list:
                raise TypeError
             
            allstr_self = all(type(x) == str for x in self.container)
            allint_iterable = all(type(x) == int for x in iterable)
            
            # [1, 2, 3] * ["s1", "s2", "s3"]
            if allstr_self and allint_iterable:
                return List(list(map(lambda x: x[0] * x[1], zip(self.container, iterable))))
            else:
                allint_self = all(type(x) == int for x in self.container)
                allstr_iterable = all(type(x) == str for x in iterable)
                
                # ["s1", "s2", "s3"] * [1, 2, 3]
                if allint_self and allstr_iterable:
                    return List([tp[0] * tp[1] for tp in zip(self.container, iterable)])
                else:
                    allreal_self = all(type(x) == int or type(x) == float for x in self.container)
                    allreal_instance = all(type(x) == int or type(x) == float for x in self.container)

                    if allreal_self and allreal_instance:
                        return List([tp[0] * tp[1] for tp in zip(self.container, iterable)])
                    else:
                        raise ValueError("cant multiply non ints")


    def __div(self, instance, operator='/'):
        if self.container == []:
            raise EmptyListError
        
        instance_type = type(instance)
        if instance_type == int or instance_type == float:
            allreal_self = all(type(x) == int or type(x) == float for x in self.container)
                
            if allreal_self:
                if operator == '//':
                    return List([x // instance for x in self.container])
                return List([x / instance for x in self.container])
            else:
                raise ValueError("cant floor div non ints")
        else:
            iterable = instance
            if type(instance) == List:
                iterable = instance.container
            elif type(iterable) == tuple:
                pass
            elif type(instance) != list:
                raise TypeError
            
            allreal_self = all(type(x) == int or type(x) == float for x in self.container)
            allreal_iterable = all(type(x) == int or type(x) == float for x in self.container)
            
            if allreal_self and allreal_iterable:
                if operator == '//':
                    return List([tp[0] // tp[1] for tp in zip(self.container, iterable)])
                return List([tp[0] / tp[1] for tp in zip(self.container, iterable)])
            else:
                raise ValueError("cant floor div non ints by non ints")


    def __div__(self, instance):
        """ Dividing every int element of the 
                self.container with instance.
            Or dividing every int or float elem
                of the self.container with every 
                element of the instance. 
        """
        return self.__div(instance)


    def __floordiv__(self, instance):
        """ Dividing every int element of the 
                self.container with instance.
            Or dividing every int or float elem
                of the self.container with every 
                element of the instance. 
        """
        return self.__div(instance, operator='//')


    def __mod__(self):
        raise NotImplementedError


    def __pow__(self):
        raise NotImplementedError


    def __rshift__(self, instance):
        if self.container == []:
            raise EmptyListError
        
        instance_type = type(instance)
        if instance_type != int:
            raise TypeError
        if instance < 0:
            raise ValueError
        result = self.container.copy()
        
        if instance == 0:
            return List(result)
        times = instance % len(self.container)
        if times == 0:
            return List(result)
            
        for _ in range(times):
            last = result[len(result) - 1]
            result = [last] + result[:len(result) - 1]
        return List(result)


    def __lshift__(self, instance):
        if self.container == []:
            raise EmptyListError
        
        instance_type = type(instance)
        if instance_type != int:
            raise TypeError
        if instance < 0:
            raise ValueError
        
        result = self.container.copy()
        if instance == 0:
            return List(result)
        times = instance % len(self.container)
        if times == 0:
            return List(result)
            
        for _ in range(times):
            first = result[0]
            result = result[1: ] + [first]
        return List(result)
    
    # common elements from both lists
    def __and__(self, instance):
        iterable = instance
        if type(instance) == List:
            iterable = instance.container
        elif type(iterable) == tuple:
            pass
        elif type(instance) != list:
            raise TypeError
            
        return List([x for x in self.container if x in iterable])

    # simmetric difference    
    def __xor__(self, instance):
        iterable = instance
        if type(instance) == List:
            iterable = instance.container
        elif type(iterable) == tuple:
            pass
        elif type(instance) != list:
            raise TypeError
        
        from_self = List([x for x in self.container if x not in iterable])
        from_iterable = List([x for x in iterable if x not in self.container])
        return from_self + from_iterable


    def __or__(self, instance):
        raise NotImplementedError


    # plus equal (+=)
    def __iadd__(self, instance):
        self = self + instance
        return self


    # minus equal (-=)
    def __isub__(self, instance):
        self = self - instance
        return self


    def __imul__(self, instance):
        self = self * instance
        return self


    def __idiv__(self, instance):
        self = self / instance
        return self


    def __ifloordiv__(self, instance):
        self = self // instance
        return self


    def __imod__(self, instance):
        self = self % instance
        return self
    
    def __pow__(self, instance):
        self = self ** instance
        return self


    def __irshift__(self, instance):
        self = self >> instance
        return self


    def __ilshift__(self, instance):
        self = self << instance
        return self


    def __iand__(self, instance):
        self = self & instance
        return self


    def __ixor__(self, instance):
        self = self ^ instance
        return self


    def __ior__(self, instance):
        self = self | instance
        return self


    def __filter(self, operator='>'):
        result = []
        for x in self.container:
            if (type(x) == int or type(x) == float):
                if operator == '>':
                    if x > 0:
                        # making negatives from positives
                        result.append(-x)
                    else:
                        result.append(x)
                elif operator == '<':
                    if x < 0:
                        # making positives from negatives
                        result.append(-x)
                    else:    
                        result.append(x)
                else:
                    raise TypeError("bad operator...")
            else:
                result.append(x)
        return List(result)


    def __neg__(self):
        """ Negative version of all positive int elements.
            If non int elements are existent then 
                we do nothing to them.
            Does not modify in-place!
        """
        return self.__filter()


    def __pos__(self):
        """ Positive version of all negative int elements.
            If non int elements are existent then 
                we do nothing to them.
            Does not modify in-place!
        """
        return self.__filter(operator='<')


    def __abs__(self):
        return self.__pos__()


    # applies ~ to every integer in the self.container
    def __invert__(self):
        for index in range(len(self.container)):
            if type(self.container[index]) == int:
                self.container[index] = ~self.container[index]
        return self


    def __complex__(self):
        # for index in range(len(self.container)):
        #     if type(self.container[index]) == int or \
        #        type(self.container[index]) == float:
        #         self.container[index] = complex(self.container[index])
        # return self
        raise NotImplementedError


    def __int__(self):
        raise NotImplementedError


    def __long__(self):
        raise NotImplementedError


    def __float__(self):
        raise NotImplementedError


    def __oct__(self):
        raise NotImplementedError


    def __hex__(self):
        raise NotImplementedError


    def __iter__(self):
        return iter(self.container)


    def __dir__(self):
        return dir(List)


    def sorted(self, *, key=None, reverse=False):
        return sorted(self.container, key=key, reverse=reverse)


    def type(self):
        return "<class 'List'>"


    def __contains__(self, item):
        """ 'in' operator """
        if self.container == []:
            raise EmptyListError
        return item in self.container


    def __delattr__(self, name):
        if name == 'container':
            raise PermissionError("How dare you delete the best list ever!?")
        else:
            print("eh, that dimension was useless...")
            object.__delattr__(self, name)


    def __hash__(self):
        pass


if __name__ == '__main__':
    ultra_list = List("12312312")
    ultra_list.append(123)
    ultra_list.append_front(random.randrange(0, 10))
    print(ultra_list.count(123))
    print(ultra_list.last_item())
    print(ultra_list.__dict__)
    print(ultra_list.)
    print(ultra_list)



"""
    core/__path.py

    useful module in development of programs
    that work with paths

    author: @alexzander
"""


# python
import os
from time import sleep
from string import ascii_lowercase

# core package (pip install python-core)
from core.system import *
from core.aesthetics import *
from core.exceptions import *

# 3d party
from win32com.client import Dispatch # pip install pywin32


# no matter what os you have
# / (slash) is the best on all platforms
# and also
# op_sys = get_os()
# if op_sys == "Windows":
#     separator__ = "\\"
separator__ = "/"
# SEPARATOR \\ IS FORBIDDEN !!!!


win_drive_letters__ = [letter + ":" for letter in ascii_lowercase]
illegal_chars__ = ":*?\"<>|"


class __path_Exception(Exception):
    """
        __path (name of this file)
        _Exception
        means this is a exception super class
        for this file only
    """
    def __init__(self, message=""):
        self.message = message


class InvalidPathError(__path_Exception):
    pass


class InvalidExtensionError(__path_Exception):
    pass


class IncorrectExtensionError(__path_Exception):
    pass


class IllegalPathError(__path_Exception):
    pass


class DifferentExtensionsError(__path_Exception):
    pass


class ContainsNoSeparatorError(__path_Exception):
    pass


class NotAFileError(__path_Exception):
    pass


class NotAFolderError(__path_Exception):
    pass


def get_path_sep(path: str):
    """ return path separator (\\ or /) from @path """
    back_slashes = path.count("\\")
    slashes = path.count("/")
    if (back_slashes == 0 and slashes > 0) or (back_slashes > 0 and slashes > 0):
        return "/"
    elif back_slashes > 0 and slashes == 0:
        return "\\"
    elif back_slashes == 0 and slashes == 0:
        raise IllegalPathError("path: {} doesnt contain any (\\) or (/)".format(red_bold(path)))


def is_valid_path(path: str):
    """ shouldnt exist, must contain separators """
    op_sys = get_os()
    if op_sys == "Windows":
        back_slashes = path.count("\\")
        slashes = path.count("/")
        if back_slashes == 0 and slashes == 0:
            # doesnt contain path separator
            return False

        # beginning = path[:2].lower() # (C:) only
        # if beginning not in drive_letters__:
        #     # doesnt start with valid windows drive letter
        #     return False
        # if path[3] != "\\" and path[3] != "/":
        #     # doesnt start with separator after drive letter
        #     return False
        for char in path[4:]:
            if char in illegal_chars__:
                return False

    elif op_sys == "Linux" or op_sys == "Darwin":
        slashes = path.count("/")
        if slashes == 0:
            return False
        for char in path:
            if char in illegal_chars__:
                return False

    return True


def get_path_from_absolute(path: str):
    """ extracts the path from @src

        "folder1/folder2/folder3/folder4/file_name.extension" =>
        => "folder1/folder2/folder3/folder4"
    """

    # validation
    if type(path) != str:
        raise TypeError
    if not is_valid_path(path):
        raise InvalidPathError
    # /validation

    if get_path_sep(path) == "\\":
        return separator__.join(path.split("\\")[:-1])
    return separator__.join(path.split(separator__)[:-1])


def is_file(path: str):
    """ should exist on disk and should be a avalid file path format """
    return os.path.isfile(path)


def is_folder(path: str):
    """ should exist on disk and should be a valid folder path format """
    return os.path.isdir(path)


def is_folder_empty(path: str):
    """ tests whether the @source_folder is empty or not """

    # validation
    if type(path) != str:
        path = str(path)
    if not is_folder(path):
        raise NotAFolderError
    # /validation
    return os.listdir(path) == 0


def is_file_empty(path: str):
    if not is_file(path):
        raise NotAFileError

    try:
        content = open(path, "r+").read()
        if content == "ÿþ" or content == "":
            return True
    except Exception as error:
        print(error)
        print(error.message)
        print(type(error))

    return False


def is_abs(path: str):
    """ doesnt need to exist and should be a valid folder or file path format """
    return os.path.isabs(path)


def exists(path: str):
    """ should exist and should be a valid folder folder or file path format """
    return os.path.exists(path)


def delete_last_slash(path: str):
    """ gets [
            'folder1\folder2\folder3\'
                        or
            'folder1/folder2/folder3/'
        ]
        return 'folder1\folder2\folder3
    """
    if path.endswith("\\") or path.endswith("/"):
        # modification through reference
        path = path[:-1]


def get_file_name(path: str):
    """ gets 'folder1/folder2/folder3/file_name.extension"
        return file_name
    """

    # validation
    if type(path) != str:
        path = str(path)
    if not is_valid_path(path):
        raise InvalidPathError
    # /validation

    sep = get_path_sep(path)

    if "." not in path:
        return path.split(sep)[-1]

    # noisnetxe.eman_elif (file_name.extension revered)
    # it can contain multiple '.' (dots) and that is confusing
    full_file_name_rev = path.split(sep)[-1][::-1]

    # from the first '.' to the end, reversed
    return full_file_name_rev[full_file_name_rev.index(".") + 1: ][::-1]


def get_file_extension(path: str):
    """ gets 'folder1/folder2/folder3/folder4/file_name.extension'
        return 'extension'
    """

    # validation
    if type(path) != str:
        path = str(path)
    if not is_valid_path(path):
        raise InvalidPathError
    if "." not in path:
        raise ValueError("there is no '.' (dot) in the @path - [this file has no extension]")
    # /validation

    # noisnetxe.eman_elif (file_name.extension reversed)
    path_rev = path[::-1]
    i = path_rev.find(".")
    return path_rev[:i][::-1]


def get_filename_plus_extension(path: str):
    """ gets 'folder1/folder2/folder3/folder4/file_name.extension'
        return 'file_name.extension'
    """

    # validation
    if type(path) != str:
        path = str(path)
    if not is_valid_path(path):
        raise InvalidPathError
    # /validation

    return path.split(get_path_sep(path))[-1]


from core.drive import *


class File:
    def __init__(self, path: str):
        if not is_valid_path(path):
            raise InvalidPathError

        self.path = path
        self.cwd = get_path_from_absolute(path)
        self.name = get_file_name(path)
        self.extension = get_file_extension(path)
        self.name_plus_extension = get_filename_plus_extension(path)
        self.sep = get_path_sep(path)

        if is_file(path):
            self.size_in_bytes = get_size_in_bytes(path)


class Folder:
    def __init__(self, path: str):
        if not is_valid_path(path):
            raise InvalidPathError

        self.path = path
        self.cwd = get_path_from_absolute(path)
        self.sep = get_path_sep(path)
        self.name = path.split(sep)[-1]

        if is_folder(path): # exists on disk
            self.size_in_bytes = get_size_in_bytes(path)


space_4  = " " * 4
branch = '│   '
tee    = '├───'
final  = '└───'

def tree_repr(folder: str, prefix=" " * 4):
    representation = ""

    __items = os.listdir(folder)
    pointers = [tee] * (len(__items) - 1) + [final]

    sep = get_path_sep(folder)
    for pointer, __item in zip(pointers, __items):
        full_path = folder + sep + __item
        representation += prefix + pointer + __item + "\n"
        if is_folder(full_path):
            extension = branch if pointer == tee else space_4
            representation += tree_repr(full_path, prefix + extension)

    return representation


def tree_representation(folder: str):
    if not is_folder(folder):
        raise NotAFolderError

    sep = get_path_sep(folder)
    result = tree_repr(folder)
    result = folder.split(sep)[-1] + " ( {} )\n".format(folder) + result
    return result




# TESTING
if __name__ == '__main__':
    pass




# core package (pip install python-core)
from core.system import get_os

op_sys = get_os()

sounds_remote_path = None
if op_sys == "Windows":
    sounds_remote_path = "C:/Users/{username}/sounds/{type}"
elif op_sys == "Linux":
    sounds_remote_path = "/home/{username}/sounds/{type}"
elif op_sys == "Darwin":
    sounds_remote_path = "/Users/{username}/sounds/{type}"

del op_sys